/******************************************************************************
* C Source: Field                                                             
*                                                                             
* File:                                                                       
*    field.c                                                                  
*                                                                             
* Description:                                                                
*    This set of routines manipulate a structure (see field.h) which contains 
*    data about a user requested field, namely, it's name, center right       
*    ascension and declination in radians, and width and height in            
*    arcminutes.                                                              
*                                                                             
* Module List:                                                                
*    Compare_Fields :                                                         
*       Lexigraphically compares data items of two fields.                    
*    Create_Field   :                                                         
*       Create memory for a field and inserts its data.                       
*    Delete_Field   :                                                         
*       Free memory for a field.                                              
*    Read_Field_Info:                                                         
*       Read data for a single field from a file.                             
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 28-SEP-95 by J. Doggett,                                    
*       Updated Read_Field_Info function.                                     
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
*                                                                             
* Copyright (c) 1993-1996, Association of Universities for Research in        
* Astronomy, Inc.  All Rights Reserved.                                       
* Produced under National Aeronautics and Space Administration grant          
* NAG W-2166.                                                                 
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "field.h"

#include "astronmy.h"
#include "generic.h"

/******************************************************************************
* Module:                                                                     
*    Compare_Fields                                                           
*                                                                             
* File:                                                                       
*    field.c                                                                  
*                                                                             
* Description:                                                                
*    Compares the data of two fields, in the order of name, right ascension,  
*    declination, width and height.                                           
*                                                                             
* Return Value:                                                               
*    int: The result of the comparison of the first data item that is         
*       different. If all the items are equal, then zero is returned.         
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Field_1                                                               
*          FIELD, Reference                                                   
*             The first field to compare.                                     
*       Field_2                                                               
*          FIELD, Reference                                                   
*             The second Field to compare.                                    
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/******************************
* Compare_Fields declarations *
******************************/
int   Compare_Fields( Field_1, Field_2 )
FIELD Field_1;
FIELD Field_2;
{
   int Comparison; /* Result of the comparison of a data item. */

/***************************************
* Compare_Fields begin executable code *
***************************************/
   Comparison = strcmp( Field_1->Name, Field_2->Name );
   if ( Comparison != 0 )
      return Comparison;

   Comparison = ( int )( Field_1->Center_RA - Field_2->Center_RA );
   if ( Comparison != 0 )
      return Comparison;

   Comparison = ( int )( Field_1->Center_Dec - Field_2->Center_Dec );
   if ( Comparison != 0 )
      return Comparison;

   Comparison = ( int )( Field_1->Width - Field_2->Width );
   if ( Comparison != 0 )

/* *** Compare_Fields ***                                                 2 */


      return Comparison;

   Comparison = ( int )( Field_1->Height - Field_2->Height );
   if ( Comparison != 0 )
      return Comparison;

   return ( int )0;
}
/*************************************
* Compare_Fields end executable code *
*************************************/

/******************************************************************************
* Module:                                                                     
*    Create_Field                                                             
*                                                                             
* File:                                                                       
*    field.c                                                                  
*                                                                             
* Description:                                                                
*    Allocates memory to hold data for a field, and inserts the data into the 
*    memory.                                                                  
*                                                                             
* Return Value:                                                               
*    FIELD: The newly created Field.                                          
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Name                                                                  
*          STRING, Reference                                                  
*             Name for the field.                                             
*       Center_RA                                                             
*          double, Value                                                      
*             The right ascension of the field center in radians.             
*       Center_Dec                                                            
*          double, Value                                                      
*             The declination of the field center in radians.                 
*       Width                                                                 
*          double, Value                                                      
*             Width of the field in arcminutes.                               
*       Height                                                                
*          int, Value                                                         
*             Height of the field in arcminutes.                              
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/****************************
* Create_Field declarations *
****************************/
FIELD  Create_Field( Name, Center_RA, Center_Dec, Width, Height )
STRING Name      ;
double Center_RA ;
double Center_Dec;
double Width     ;
double Height    ;
{
   FIELD New_Field; /* Newly creately field. */

/*************************************
* Create_Field begin executable code *
*************************************/
   New_Field = ( FIELD )Allocate_Memory( sizeof( *New_Field ) );

   strcpy( New_Field->Name, Name )   ;

/* *** Create_Field ***                                                   2 */


   New_Field->Center_RA  = Center_RA ;
   New_Field->Center_Dec = Center_Dec;
   New_Field->Width      = Width     ;
   New_Field->Height     = Height    ;

   return New_Field;
}
/***********************************
* Create_Field end executable code *
***********************************/

/******************************************************************************
*                                                                             
* Module:                                                                     
*    Delete_Field                                                             
*                                                                             
* File:                                                                       
*    field.c                                                                  
*                                                                             
* Description:                                                                
*    Free memory previously allocated for a field.                            
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Field                                                                 
*          FIELD, Reference                                                   
*             The field we're deleting from memory.                           
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/****************************
* Delete_Field declarations *
****************************/
void  Delete_Field( Field )
FIELD Field;
{
/*************************************
* Delete_Field begin executable code *
*************************************/
   if ( Field != NULL )
      free( Field );

   return;
}
/***********************************
* Delete_Field end executable code *
***********************************/

/******************************************************************************
* Module:                                                                     
*    Read_Field_Info                                                          
*                                                                             
* File:                                                                       
*    field.c                                                                  
*                                                                             
* Description:                                                                
*    Read data for field from a file.                                         
*                                                                             
* Return Value:                                                               
*    FIELD: The field read from the file.                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Equinox                                                               
*          EQUINOX, Value                                                     
*             Equinox of the coordinates read in for the field.               
*       InputFile                                                             
*          FILE*, Value                                                       
*             File pointer to the input file.                                 
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 28-SEP-95 by J. Doggett,                                    
*       Modified to reflect increase in limit of field name size.             
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
******************************************************************************/
/*******************************
* Read_Field_Info declarations *
*******************************/
FIELD Read_Field_Info( Equinox, InputFile )
EQUINOX Equinox  ;
FILE*   InputFile;
{
   FIELD Field; /* Place for field data read in */

/* Variables for reading data */
   char   Field_Name[FIELD_NAME_MAX_LENGTH + 1]
                         ; /* characters; must be valid for a filename */
   int    RA_Hours       ; /* Field center                       */
   int    RA_Minutes     ;
   double RA_Seconds     ;
   char   Dec_Sign       ;
   int    Dec_Degrees    ;
   int    Dec_ArcMinutes ;
   double Dec_ArcSeconds ;
   double RA_Radians     ;
   double Dec_Radians    ;
   double Field_Width    ; /* Field size in arcminutes           */
   double Field_Height   ; /* Field size in arcminutes           */

/* *** Read_Field_Info ***                                                2 */


   int    Temp_Char,sret      ;

/****************************************
* Read_Field_Info begin executable code *
****************************************/
/* Read field name */
   sret=fscanf(InputFile, "%s", Field_Name);
   switch(sret)
   {
      case 1:
         break;
      case EOF:
         return ( FIELD )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != '\n')
                                   &&
                (Temp_Char                     != EOF )
               ) ;
         fprintf(stderr,"Error reading field name.\n");
         return ( FIELD )NULL;
   }
   Field_Name[FIELD_NAME_MAX_LENGTH] = (char)NULL;

/* Read field position */
   switch( Read_DMS( InputFile, &RA_Radians, &Dec_Radians ) )
   {
      case 1:
         break;
      case EOF:
         return ( FIELD )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != '\n')
                                   &&
                (Temp_Char                    != EOF )
               ) ;
         fprintf(stderr,"Error with input field coordinates.\n");
         return ( FIELD )NULL;
   }

/* Read field size */
   switch( fscanf( InputFile, "%lf %lf",
                   &Field_Width, &Field_Height ) )
   {
      case 2:
         while (( (Temp_Char=getc(InputFile)) != '\n')
                            &&
                (Temp_Char                    != EOF )
               ) ;
         if ( Equinox == B1950 )
            b1950_j2000(  RA_Radians, Dec_Radians, &RA_Radians, &Dec_Radians  );
         Field = Create_Field
            ( Field_Name, RA_Radians, Dec_Radians, Field_Width, Field_Height );
         return Field;
      case EOF:
         return ( FIELD )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != '\n')
                            &&
                (Temp_Char                    != EOF )
               ) ;
         fprintf(stderr,"Error reading field information.\n");
         return ( FIELD )NULL;

/* *** Read_Field_Info ***                                                3 */


   }
}
/**************************************
* Read_Field_Info end executable code *
**************************************/
