/******************************************************************************
* C Source: Header                                                            
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    This set of functions are used for processing the header information for 
*    a plate (see header.h).  These functions were either originally          
*    written, or ported to C by R. White unless otherwise noted.              
*                                                                             
* Module List:                                                                
*    Check_Header_Plate_Id                                                    
*       Compares the plate id of header against input plate id using strcmp.  
*    Delete_Header                                                            
*       Delete memory for header.                                             
*    fillhdr                                                                  
*       Change FITS header parameters to match an image section's parameters. 
*    fillpos                                                                  
*       Add FITS header parameters for RA, DEC for an object.                 
*    gethdr                                                                   
*       Organizes the read of a a FITS header into a data structure.          
*    Make_Header_File_Name                                                    
*       Create full header file name given header directory, and plate name.  
*    readhdr                                                                  
*       Reads a FITS header into a data structure.                            
*                                                                             
* History:                                                                    
*    Created    : 13-AUG-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 20-JUL-94 by J. Doggett,                                    
*       Updated gethdr function.                                              
*    Updated    : 30-SEP-94 by J. Doggett,                                    
*       Updated fillhdr function.                                             
*    Redelivered: 30-NOV-94 by J. Doggett,                                    
*       GetImage Version 1.1.                                                 
*                                                                             
*    Updated    : 10-JAN-95 by J. Doggett,                                    
*       Updated fillhdr function.                                             
*    Updated    : 07-APR-95 by J. Doggett,                                    
*       Updated readhdr function.                                             
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
*                                                                             
* Copyright (c) 1993-1995, Association of Universities for Research in        
* Astronomy, Inc.  All Rights Reserved.                                       
* Produced under National Aeronautics and Space Administration grant          
* NAG W-2166.                                                                 
******************************************************************************/
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "header.h"
#include "fitsio.h"
#include "fitsio2.h"
/*****************************
* Header introduction page 2 *
*****************************/

#include "astronmy.h"
#include "config.h"
#include "filenams.h"
#include "generic.h"
/*#include "dss2wcs.h"*/

/******************************************************************************
* Module:                                                                     
*    Check_Header_Plate_Id                                                    
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Compares the plate id of header against input plate id using strcmp.     
*                                                                             
* Return Value:                                                               
*    int: The comparison result returned from strcmp.                         
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Header                                                                
*          HEADER, Reference                                                  
*             Header to check against input plate id.                         
*       Plate_Id                                                              
*          STRING, Reference                                                  
*             Plate id we're checking for.                                    
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 23-AUG-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/*************************************
* Check_Header_Plate_Id declarations *
*************************************/
int    Check_Header_Plate_Id( Header, Plate_Id )
HEADER Header  ;
STRING Plate_Id;
{
/**********************************************
* Check_Header_Plate_Id begin executable code *
**********************************************/
   return strcmp( Header->plate_id, Plate_Id );
}
/********************************************
* Check_Header_Plate_Id end executable code *
********************************************/

/******************************************************************************
* Module:                                                                     
*    Delete_Header                                                            
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Free memory previously allocated for a header.                           
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Header                                                                
*          HEADER, Reference                                                  
*             The header we're deleting from memory.                          
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 02-SEP-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/*****************************
* Delete_Header declarations *
*****************************/
void  Delete_Header( Header )
HEADER Header;
{
/**************************************
* Delete_Header begin executable code *
**************************************/
   if ( Header != NULL )
      free( Header );

   return;
}
/************************************
* Delete_Header end executable code *
************************************/

/******************************************************************************
* Module:                                                                     
*    fillhdr                                                                  
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*       Change FITS header parameters to match an image sections parameters.  
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input/Output Arguments:                                                  
*       h                                                                     
*          HEADER, Reference                                                  
*             The header to be modified.                                      
*    Input Arguments:                                                         
*       object                                                                
*          char, Reference                                                    
*             String containg the object's name.                              
*       nx                                                                    
*          int, Value                                                         
*             Width of the image.                                             
*       ny                                                                    
*          int, Value                                                         
*             Height of the image.                                            
*       x0                                                                    
*          int, Value                                                         
*             X of the lower left corner of the image on the plate.           
*       y0                                                                    
*          int, Value                                                         
*             Y of the lower left corner of the image on the plate.           
*       datamin                                                               
*          int, Value                                                         
*             Minimum data value in the image.                                
*       datamax                                                               
*          int, Value                                                         
*             Maximum data value in the image.                                
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 31-JUL-91 by R. White,                                      
*    Delivered  : 31-JUL-91 by R. White,                                      
*                                                                             
*    Updated    : 13-AUG-93 by J. Doggett,                                    
*       Adopted into software used to facilitate the accessing of data from   
*          the CD-ROM set of the Digitized Sky Survey.                        
*    Updated    : 08-DEC-93 by J. Doggett,                                    
*       Added option for not computing DATAMIN and DATAMAX when input image   
*          pointer is NULL.                                                   
*    Updated    : 14-DEC-93 by J. Doggett,                                    
*       Modified to received minimum and maximum data values of the image     
*          input parameters, rather determining directly from the image.      
*    Updated    : 03-FEB-94 by J. Doggett,                                    
*       Modified to use keyword CNPIX rather than CRPIX.                      
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 30-SEP-94 by J. Doggett,                                    
******************************************************************************
									      
******************							      
* fillhdr page 2 *							      
******************							      
									      
******************************************************************************
*       Modified to create DATE conforming to "FITS" standard.                
*    Redelivered: 30-NOV-94 by J. Doggett,                                    
*       GetImage Version 1.1.                                                 
*                                                                             
*    Updated    : 10-JAN-95 by J. Doggett,                                    
*       Fixed bug in adding CNPIX1/2 keywords if not already present in the   
*       input header.                                                         
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
******************************************************************************/
extern void
fillhdr(h,object,nx,ny,x0,y0,datamin,datamax)

HEADER h;
char *object;
int nx,ny,x0,y0;
int datamin, datamax;
{
int i,j;
char keyword[9], hlast[81];
time_t itime;
time_t timer;
long	inaxes[2], incorner[2];
int		ierr;
char *cdate, caltype, date[18];
int no_naxis1, no_naxis2, no_cnpix1, no_cnpix2, no_datamax, no_datamin,
    no_object, no_date, wcsfive;
double testx,testy,testra,testdec;


    /*
     * set size of section and corner position
     */
    h->section_x_length = nx;
    h->section_y_length = ny;
    h->section_x_corner = x0;
    h->section_y_corner = y0;
    /*
     * object name, padded with blanks
     */
    for (i=0; i<sizeof(h->object_name)-1; i++) h->object_name[i] = ' ';
    h->object_name[i] = '\0';
    for (i=0; (i<sizeof(h->object_name)-1) && (object[i] != '\0'); i++)
        h->object_name[i] = object[i];
    /*
     * get date & time
     * original format is 'wkd mmm dd hh:mm:ss yyyy\n\0'
     * change to 'mmm dd hh:mm yyyy\0' so it fits in <18 characters
     */
    itime = time(NULL);
    strftime(date, 18, "%Y-%m-%d", localtime(&itime));
    /*
     * look for existing keywords and change them
     */
    no_naxis1  = 1;
    no_naxis2  = 1;
    no_cnpix1  = 1;
    no_cnpix2  = 1;
    no_datamax = 1;
    no_datamin = 1;
    no_object  = 1;

    /*
     * Add WCS to Header
     */
	 if (h->amdre_flag =='T') caltype='2'; else caltype='1';
	 ComputeWCS(h,caltype);

	/* inaxes[0]=h->section_x_length;
	incorner[0]=h->section_x_corner;
	inaxes[1]=h->section_y_length;
	incorner[1]=h->section_y_corner;
	strncpy(h->ctype[0] ,"RA---TAN",8) ;
	h->ctype[0][8]='\0';
	strncpy(h->ctype[1] ,"DEC--TAN",8) ;
	h->ctype[1][8]='\0';
	dss2wcs (inaxes,incorner,h->ppo_coeff,h->plt_scale,h->x_pixel_size,h->y_pixel_size,
			h->plt_center_ra*57.2938,h->plt_center_dec*57.2938,
			h->ppo_coeff[2],h->ppo_coeff[5], h->amd_x_coeff,h->amd_y_coeff,
			(char**)&h->ctype,(float*)&h->cdelt,(float*)&h->crpix,(double*)&h->crval,(float*)&h->crota,
			&ierr);
	if (ierr > 0) fprintf(stdout,"*************** DSS2WCS error ************** "); */
  

/*****************
* fillhdr page 3 *
*****************/

    no_date    = 1;

    for (i=0; i < h->nhlines; i++) {
        /*
         *  Extract keyword name and put in null-terminated string
         */
        strncpy(keyword, h->hlines[i], 8);
        keyword[8] = '\0';
        /*
         *  Find keywords to change
         */
        if        (strcmp(keyword,"NAXIS1  ") == 0) {
            sprintf(&h->hlines[i][10],"%20d",h->section_x_length);
            h->hlines[i][30] = ' ';
            no_naxis1 = 0;
        } else if (strcmp(keyword,"NAXIS2  ") == 0) {
            sprintf(&h->hlines[i][10],"%20d",h->section_y_length);
            h->hlines[i][30] = ' ';
            no_naxis2 = 0;
        } else if (strcmp(keyword,"CNPIX1  ") == 0) {
            sprintf(&h->hlines[i][10],"%20.0f",h->section_x_corner);
            h->hlines[i][30] = ' ';
            no_cnpix1 = 0;
        } else if (strcmp(keyword,"CNPIX2  ") == 0) {
            sprintf(&h->hlines[i][10],"%20.0f",h->section_y_corner);
            h->hlines[i][30] = ' ';
            no_cnpix2 = 0;
        } else if (strcmp(keyword,"DATAMAX ") == 0) {
            sprintf(&h->hlines[i][10],"%20d",datamax);
            h->hlines[i][30] = ' ';
            no_datamax = 0;
        } else if (strcmp(keyword,"DATAMIN ") == 0) {
            sprintf(&h->hlines[i][10],"%20d",datamin);
            h->hlines[i][30] = ' ';
            no_datamin = 0;
        } else if (strcmp(keyword,"OBJECT  ") == 0) {
            sprintf(&h->hlines[i][10],"'%-18.18s'",h->object_name);
            h->hlines[i][30] = ' ';
            no_object = 0;
        } else if (strcmp(keyword,"DATE    ") == 0) {
            sprintf(&h->hlines[i][10],"'%-18.18s'",date);
            h->hlines[i][30] = ' ';
            no_date = 0;
        }
    }
    /*
     * if some of these keywords were absent, insert them just before
     *  the last line of the header
     */
    i = h->nhlines-1;
    for (j=0; j<sizeof(hlast); j++) hlast[j] = h->hlines[i][j];

    if (no_naxis1) {
        sprintf(h->hlines[i], "%-8.8s= %20d /%-50.50s",
            "NAXIS1", h->section_x_length, "Length X axis");
        i = i + 1;
    }
    if (no_naxis2) {
        sprintf(h->hlines[i], "%-8.8s= %20d /%-50.50s",
            "NAXIS2", h->section_y_length, "Length Y axis");

/*****************
* fillhdr page 4 *
*****************/

        i = i + 1;
    }
    if (no_cnpix1) {
        sprintf(h->hlines[i], "%-8.8s= %20.0f /%-50.50s",
            "CNPIX1", h->section_x_corner, "X corner");
        i = i + 1;
    }
    if (no_cnpix2) {
        sprintf(h->hlines[i], "%-8.8s= %20.0f /%-50.50s",
            "CNPIX2", h->section_y_corner, "Y corner");
        i = i + 1;
    }
    if (no_date) {
        sprintf(h->hlines[i], "%-8.8s= '%-18.18s' /%-50.50s",
            "DATE", date, "Creation Date");
        i = i + 1;
    }

	/* Adding WCS Keywords */
    sprintf(h->hlines[i],"WCSAXES = %20i /%-50.50s", h->wcsaxes, "GetImage: Number WCS axes");
    i++;
    sprintf(h->hlines[i],"WCSNAME = '%-18.18s' /%-50.50s", h->wcsname, "GetImage: Local WCS approximation from full plate solution");
    i++;
	sprintf(h->hlines[i],"RADESYS = '%-18.18s' /%-50.50s", h->radesys, "GetImage: GSC-II calibration using ICRS system");
    i++;
    sprintf(h->hlines[i],"CTYPE1  = '%-18.18s' /%-50.50s", h->ctype[0], "GetImage: RA-Gnomic projection");
    i++;
	sprintf(h->hlines[i],"CRPIX1  = %20.6f /%-50.50s", h->crpix[0], "GetImage: X reference pixel");
    i++;
	sprintf(h->hlines[i],"CRVAL1  = %20.6f /%-50.50s", h->crval[0], "GetImage: RA of reference pixel");
    i++;
	sprintf(h->hlines[i],"CUNIT1  = '%-18.18s' /%-50.50s", h->cunit[0], "GetImage: degrees");
    i++;
    sprintf(h->hlines[i],"CTYPE2  = '%-18.18s' /%-50.50s", h->ctype[1], "GetImage: Dec-Gnomic projection");
	i++;
	sprintf(h->hlines[i],"CRPIX2  = %20.6f /%-50.50s", h->crpix[1], "GetImage: Y reference pixel");
    i++;
	sprintf(h->hlines[i],"CRVAL2  = %20.6f /%-50.50s", h->crval[1], "GetImage: Dec of reference pixel");
    i++;
	sprintf(h->hlines[i],"CUNIT2  = '%-18.18s' /%-50.50s", h->cunit[1], "Getimage: degrees");
    i++;
	sprintf(h->hlines[i],"CD1_1   = %20.10f /%-50.50s", h->cdmatrix[0], "GetImage: rotation matrix coefficient");
    i++;
	sprintf(h->hlines[i],"CD1_2   = %20.10f /%-50.50s", h->cdmatrix[1], "GetImage: rotation matrix coefficient");
    i++;
	sprintf(h->hlines[i],"CD2_1   = %20.10f /%-50.50s", h->cdmatrix[2], "GetImage: rotation matrix coefficient");
    i++;
	sprintf(h->hlines[i],"CD2_2   = %20.10f /%-50.50s", h->cdmatrix[3], "GetImage: rotation matrix coefficient");
    i++;
    sprintf(h->hlines[i],"OBJECT  = '%18.18s' /%-50.50s", h->object_name, "GetImage: Requested Object Name");
    i++;
    sprintf(h->hlines[i],"DATAMIN = %20d /%-50.50s", datamin, "GetImage: Minimum returned pixel value");
    i++;
    sprintf(h->hlines[i],"DATAMAX = %20d /%-50.50s", datamax, "GetImage: Maximum returned pixel value");
    i++;

	/* put last line back on the end*/
    if (i >= h->nhlines) {
        for (j=0; j<sizeof(hlast); j++) h->hlines[i][j] = hlast[j];
        h->nhlines = i+1;
    }

/*printf("Mask %s\n",h->mask);
printf("Refraction %f %f %f %f \n",h->ref_Beta,h->ref_BetaP,h->ref_k1,h->ref_k2);
printf("AMD %c AMDRE %c \n",h->amd_flag,h->amdre_flag);
printf("\nAdded WCS keywords\n"); 	    */

} /* fillhdr */

/******************************************************************************
* Module:                                                                     
*    fillpos                                                                  
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Add FITS header parameters for RA, Dec for this object                   
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input/Output Arguments:                                                  
*       h                                                                     
*          HEADER, Reference                                                  
*             The header to be modified.                                      
*    Input Arguments:                                                         
*       rah                                                                   
*          int, Value                                                         
*             Hours of right ascension for object                             
*       ram                                                                   
*          int, Value                                                         
*             Minutes of right ascension for object                           
*       ras                                                                   
*          double, Value                                                      
*            Seconds of right ascension for object                            
*       decsign                                                               
*          char, Value                                                        
*             Sign of declination of object                                   
*       decd                                                                  
*          int, Value                                                         
*             Degrees of declination of object                                
*       decm                                                                  
*          int, Value                                                         
*             Minutes of arc of declination of object                         
*       decs                                                                  
*          double, Value                                                      
*             Seconds of arc of declination of object                         
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 31-JUL-91 by R. White,                                      
*    Delivered  : 31-JUL-91 by R. White,                                      
*                                                                             
*    Updated    : 31-AUG-93 by J. Doggett,                                    
*       Adopted into software used to facilitate the accessing of data from   
*          the CD-ROM set of the Digitized Sky Survey.                        
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
extern void
fillpos(h,rah,ram,ras,decsign,decd,decm,decs,x,y)
HEADER h;
int rah,ram;
double ras;
char decsign;
int decd,decm;
double decs;
double x,y;
{

/*****************
* fillpos page 2 *
*****************/

int i,j;
char keyword[9], hlast[81], rastr[13], decstr[13];
int no_ra, no_dec, no_x, no_y;
    /*
     * set position in header structure
     */
    h->ra_h = rah;
    h->ra_m = ram;
    h->ra_s = ras;
    h->dec_d = decd;
    h->dec_m = decm;
    h->dec_s = decs;
    h->dec_sign = decsign;
    h->x = x;
    h->y = y;
    /*
     * round for printing with ss.sss (RA), ss.ss (Dec)
     */
    if (ras+0.0005 >= 60.0) {
        ras = 0.0;
        ram += 1;
        if (ram == 60) {
            ram = 0;
            rah += 1;
            if (rah == 24) {
                rah = 0;
            }
        }
    }
    if (decs+0.005 >= 60.0) {
        decs = 0.0;
        decm += 1;
        if (decm == 60) {
            decm = 0;
            decd += 1;
        }
    }
    /*
     * strings for RA, Dec
     */
    sprintf(rastr, "%2.2d %2.2d %06.3f", rah, ram, ras);
    sprintf(decstr, "%c%2.2d %2.2d %05.2f", decsign, decd, decm, decs);
    /*
     * look for existing OBJCTRA, OBJCTDEC, OBJCTX, OBJCTY keywords and
     * change them
     */
    no_ra = 1;
    no_dec = 1;
    no_x = 1;
    no_y = 1;
    for (i=0; i < h->nhlines; i++) {
        /*
         *  Extract keyword name and put in null-terminated string
         */
        strncpy(keyword, h->hlines[i], 8);
        keyword[8] = '\0';
        /*
         *  Find keywords to change
         */
        if        (strcmp(keyword,"OBJCTRA ") == 0) {

/*****************
* fillpos page 3 *
*****************/

            sprintf(&h->hlines[i][10],"'%-18.18s'",rastr);
            h->hlines[i][30] = ' ';
            no_ra = 0;
        } else if (strcmp(keyword,"OBJCTDEC") == 0) {
            sprintf(&h->hlines[i][10],"'%-18.18s'",decstr);
            h->hlines[i][30] = ' ';
            no_dec = 0;
        } else if (strcmp(keyword,"OBJCTX  ") == 0) {
            sprintf(&h->hlines[i][10],"%20.2f",x);
            h->hlines[i][30] = ' ';
            no_x = 0;
        } else if (strcmp(keyword,"OBJCTY  ") == 0) {
            sprintf(&h->hlines[i][10],"%20.2f",y);
            h->hlines[i][30] = ' ';
            no_y = 0;
        }
    }
    /*
     * if some of these keywords were absent, insert them just before
     *  the last line of the header
     */
    i = h->nhlines-1;
    for (j=0; j<sizeof(hlast); j++) hlast[j] = h->hlines[i][j];
    if (no_ra) {
        sprintf(h->hlines[i], "%-8.8s= '%-18.18s' /%-50.50s",
            "OBJCTRA", rastr, "GetImage: Requested Right Ascension (J2000)");
        i = i+1;
    }
    if (no_dec) {
        sprintf(h->hlines[i], "%-8.8s= '%-18.18s' /%-50.50s",
            "OBJCTDEC", decstr, "GetImage: Requested Declination (J2000)");
        i = i+1;
    }
    if (no_x) {
        sprintf(h->hlines[i], "%-8.8s= %20.2f /%-50.50s",
            "OBJCTX", x, "GetImage: Requested X on plate (pixels)");
        i = i+1;
    }
    if (no_y) {
        sprintf(h->hlines[i], "%-8.8s= %20.2f /%-50.50s",
            "OBJCTY", y, "GetImage: Requested Y on plate (pixels)");
        i = i+1;
    }
    /*
     * put last line back on the end
     */
    if (i >= h->nhlines) {
        for (j=0; j<sizeof(hlast); j++) h->hlines[i][j] = hlast[j];
        h->nhlines = i+1;
    }
} /* fillpos */

/******************************************************************************
* Module:                                                                     
*    gethdr                                                                   
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Routine to read FITS header into structure using file name and location  
*    information already in header structure.                                 
*                                                                             
* Return Value:                                                               
*    HEADER*: Pointer to the header, or NULL if unsuccessful (usually because 
*             the header file does not exist.)                                
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Header_File_Name                                                      
*          STRING, Reference                                                  
*             Name of file with header data.                                  
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 05-AUG-91 by R. White,                                      
*    Delivered  : 05-AUG-91 by R. White,                                      
*                                                                             
*    Updated    : 27-MAY-93 by J. Doggett,                                    
*       Adopted into software used to facilitate the accessing of data from   
*          the CD-ROM set of the Digitized Sky Survey.                        
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 20-JUL-94 by J. Doggett,                                    
*       Removed Plate_Name argument which is not needed, and modified to fill 
*       the plate_name field in the header with the region_no field read from 
*       the header file.                                                      
*    Redelivered: 30-NOV-94 by J. Doggett,                                    
*       GetImage Version 1.1.                                                 
******************************************************************************/
extern HEADER gethdr(Header_File_Name)
STRING        Header_File_Name;
{
HEADER h;
FILE *infile;

    h = ( HEADER )Allocate_Memory( sizeof ( *h ) );
    if ((infile = fopen(Header_File_Name,"r"))==NULL)
       return (HEADER)NULL;

    /*
     * read header
     */
    readhdr(infile, h);
    strcpy( h->plate_name, h->region_no );
    Trim_Trailing_White_Space(h->plate_name);
    fclose(infile);
	
	return h;		
	
} /* gethdr */

/******************************************************************************
* Module:                                                                     
*    gethdrext                                                                                                                                                                                                             
*                                                                             
* Description:                                                                
*    Routine to read FITS header + extension with Astrometirc mask into structure using file name and location  
*    information already in header structure.                                 
*                                                                                                                                                        
* History:                                                                    
*    Created    : 01-Aug-2005 by B.McLean,                                                                    
*       GetImage Version 3.x.                                                                                                  
******************************************************************************/
extern HEADER gethdrext(Header_File_Name)
STRING        Header_File_Name;
{
	HEADER h;
	FILE *infile;
	int i, j, k, jj;
	fitsfile *fptr;
	int status, hdutype, hdunum, nfound, colnum, maxdim, naxis, datatype, nkeys, keypos, anynull;
	long naxes;
	char *ttype[3], *comment[80], *keyname[8];
	char card[FLEN_CARD];   /* standard string lengths defined in fitsioc.h */
	long frow      = 1;
	long felem     = 1;
	long nelem     = 1600;
	char strnull[10];
	strcpy(strnull, " ");
	float floatnull = 0.;
	
    h = ( HEADER )Allocate_Memory( sizeof ( *h ) );
    if ((infile = fopen(Header_File_Name,"r"))==NULL)
		return (HEADER)NULL;
	
/* read header */
    readhdr(infile, h);
    strcpy( h->plate_name, h->region_no );
    Trim_Trailing_White_Space(h->plate_name);
    fclose(infile);
	
/*Look for Astrometric Mask extension*/
	if (h->mask == NULL) {printf("No mask \n");}
	else {
			/*printf("Reading FITS extension %s\n",h->mask);*/
			status = 0;
			fits_open_file(&fptr, Header_File_Name, READONLY, &status);
			if (status != 0){
				fits_clear_errmsg();
				printf("Error opening FITS file %s;\n",Header_File_Name);		
				status = 0;}
/* Move to extension*/
			hdunum=2;
			fits_movabs_hdu(fptr, hdunum, &hdutype, &status);
			if (status != 0){
				fits_clear_errmsg();
				printf("Error moving to FITS Extension %s;\n",Header_File_Name);		
				status = 0;}
/* Check that it is table and read header*/			
			if (hdutype == ASCII_TBL) /*printf("\nReading ASCII table in HDU %d:\n",  hdunum)*/;
			else if (hdutype == BINARY_TBL) /*printf("\nReading binary table in HDU %d:\n", hdunum)*/;
			else{
				printf("Error: this HDU is not an ASCII or binary table\n");}
		
			fits_get_hdrpos(fptr, &nkeys, &keypos, &status);
			/*
			printf("Header listing for HDU #%d:\n", hdunum);
			for (jj = 1; jj <= nkeys; jj++)  {
				fits_read_record(fptr, jj, card, &status);
				printf("%s\n", card);}
			 */

/* Read 4 columns in table */					   
			fits_read_col(fptr, TFLOAT, 1, frow, felem, nelem, &floatnull, h->xi, &anynull, &status);
			fits_read_col(fptr, TFLOAT, 2, frow, felem, nelem, &floatnull, h->eta, &anynull, &status);
			fits_read_col(fptr, TFLOAT, 3, frow, felem, nelem, &floatnull, h->xicorr, &anynull, &status);
			fits_read_col(fptr, TFLOAT, 4, frow, felem, nelem, &floatnull, h->etacorr, &anynull, &status);
			
/* Close file when finished */			
			fits_close_file(fptr,&status);
	}

return h;		

} /* gethdrext */

/******************************************************************************
* Module:                                                                     
*    Make_Header_File_Name                                                    
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Create full header file name given header directory, and plate name.     
*                                                                             
* Return Value:                                                               
*    STRING: Name of header file.                                             
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Directory                                                             
*          STRING, Reference                                                  
*             Directory where headers reside.                                 
*       Plate_Name                                                            
*          STRING, Reference                                                  
*             Name of plate for header file.                                  
*       Header_Extension                                                      
*          STRING, Reference                                                  
*             Extension for file name.                                        
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 27-OCT-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/************************************
 Make_Header_File_Name declarations *
************************************/
STRING Make_Header_File_Name(Directory, Plate_Name, Header_Extension)
STRING Directory       ;
STRING Plate_Name      ;
STRING Header_Extension;
{
   STRING Header_File_Name     ; /* Without directory path */
   STRING Full_Header_File_Name; /* Including directory path

/**********************************************
* Make_Header_File_Name begin executable code *
**********************************************/
   Header_File_Name = (STRING)Allocate_Memory(strlen(Plate_Name)       +
                                              1                        +
                                              strlen(Header_Extension) +
                                              1 + /* For possible semcolon */
                                              1   /* For trailing null     */
                                             );

   strcpy(Header_File_Name, Plate_Name         );
   Trim_Trailing_White_Space(Header_File_Name  );
   strcat(Header_File_Name, "."                );
   strcat(Header_File_Name, Header_Extension   );

/*******************************
* Make_Header_File_Name page 2 *
*******************************/

   if (!Set_FileNames_Case(Directory, Header_File_Name)) {
      fprintf(stderr,"\nCould not find copyright file in header directory!\n");
      fprintf(stderr, "Aborting GetImage!!!\n\n");
      exit(EXIT_FAILURE);
   }

   Full_Header_File_Name =
                         Build_Full_File_Name(Directory, "", Header_File_Name);
   free(Header_File_Name);

   Header_File_Name =
         Set_FileNames_Version(Directory, Full_Header_File_Name);

   if (Header_File_Name != Full_Header_File_Name)
      if (Header_File_Name != NULL )
         free(Full_Header_File_Name);
      else {
         fprintf(stderr,
               "\nCould not find copyright file in header directory!\n");
         fprintf(stderr, "Aborting GetImage!!!\n\n");
         exit(EXIT_FAILURE);
      }

   Full_Header_File_Name = Header_File_Name;

   return Full_Header_File_Name;
}
/**********************************************
* Make_Header_File_Name begin executable code *
**********************************************/

/******************************************************************************
* Module:                                                                     
*    readhdr                                                                  
*                                                                             
* File:                                                                       
*    header.c                                                                 
*                                                                             
* Description:                                                                
*    Routine to read FITS header into structure.                              
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input Arguments:                                                         
*       infile                                                                
*          FILE*, Value                                                       
*             Pointer to the header file.                                     
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       header                                                                
*          Header, Reference                                                  
*             Structure with header information.                              
*                                                                             
* History:                                                                    
*    Created    : 04-MAY-90 by R. White,                                      
*       Converted from IDL to Fortran.                                        
*    Delivered  : 30-JUL-91 by R. White,                                      
*       Converted from Fortran to C.                                          
*                                                                             
*    Updated    : 27-MAY-93 by J. Doggett,                                    
*       Adopted into software used to facilitate the accessing of data from   
*       the CD-ROM set of the Digitized Sky Survey.                           
*    Updated    : 03-FEB-94 by J. Doggett,                                    
*       Modified to use keyword CNPIX rather than CRPIX.                      
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 07-APR-95 by J. Doggett,                                    
*       Modified to read special set of keywords containing copyright         
*       information.                                                          
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.    
*
*    Updated 02-AUG-2005 B.McLean
*	    Added GSC-II astrometry, version 3.x
******************************************************************************/
extern void
readhdr(infile,h)
FILE *infile;
HEADER h;
{
int i;
char keyword[9], value[21], line[80];
char fullCharValue[71];
int  fullCharValueLength;
int  totalNumBriefCopyRightStatements, totalNumFullCopyRightStatements;
int  currentCopyRightStatement;

    totalNumBriefCopyRightStatements = 0;
    h->numBriefCopyRightStatements   = 0;
    h->briefCopyRightStatements      = NULL;

    totalNumFullCopyRightStatements  = 0;
    h->numFullCopyRightStatements    = 0;

/*****************
* readhdr page 2 *
*****************/
    h->fullCopyRightStatements       = NULL;
    /*
     * initial plate solution coefficients to zero
     */
    for (i = 0; i<6; i++) {
        h->ppo_coeff[i] = 0.0;
    }
    for (i = 0; i<20; i++) {
        h->amd_x_coeff[i] = 0.0;
        h->amd_y_coeff[i] = 0.0;
    }
    h->nhlines = 0;
    while ((i=fread(line, 1, 80, infile)) == 80) {
        strncpy(h->hlines[h->nhlines],line,80);
        h->hlines[h->nhlines++][80] = '\0';
        /*
         *  Extract keyword name, value and put in null-terminated string
         */
        strncpy(keyword, &line[0], 8);
        keyword[8] = '\0';
        strncpy(value, &line[10], 20);
        value[20] = '\0';

        /*
         *  Get standard keywords
         */
        ival("XPIXELS ", h->section_x_length);
        ival("YPIXELS ", h->section_y_length);
        fval("CNPIX1  ", h->section_x_corner);
        fval("CNPIX2  ", h->section_y_corner);
        fval("CDELT1  ", h->scaling);
        cvalq("OBJECT  ",h->object_name);
        /*
         *  Get GSSS Specific keywords
         */
        fval("MAGLIM  ", h->mag_limit);
        ival("PLTGRADE", h->plt_grade);
        cvalt("PLATEID ",h->plate_id);
        cvalt("REGION  ",h->region_no);
        cvalt("PLTLABEL",h->plate_label);
        cvalt("EXPOSURE",h->exposure_time);
        cvalt("SEEING  ",h->seeing);
        cvalt("ORIGIN  ",h->origin);
        /*
         *  Get GSSS calibration keywords
         */
        fval("EPOCH   ", h->plt_date);
        ival("BANDPASS", h->emulsion_code);
        fval("PLTSCALE", h->plt_scale);
        fval("XPIXELSZ", h->x_pixel_size);
        fval("YPIXELSZ", h->y_pixel_size);
        ival("PLTRAH  ", h->ra_h);
        ival("PLTRAM  ", h->ra_m);
        fval("PLTRAS  ", h->ra_s);
        if (strcmp(keyword,"PLTDECSN")==0) {
            if(strchr(value,'-') == NULL) {
                h->dec_sign = '+';
            } else {
                h->dec_sign = '-';
            }
        }

/*****************
* readhdr page 3 *
*****************/

        ival("PLTDECD ", h->dec_d);
        ival("PLTDECM ", h->dec_m);
        fval("PLTDECS ", h->dec_s);
        fval("PLTVEC1 ", h->plt_center_vec[0]);
        fval("PLTVEC2 ", h->plt_center_vec[1]);
        fval("PLTVEC3 ", h->plt_center_vec[2]);
        fval("PPO1    ", h->ppo_coeff[0]);
        fval("PPO2    ", h->ppo_coeff[1]);
        fval("PPO3    ", h->ppo_coeff[2]);
        fval("PPO4    ", h->ppo_coeff[3]);
        fval("PPO5    ", h->ppo_coeff[4]);
        fval("PPO6    ", h->ppo_coeff[5]);
        fval("AMDX1   ", h->amd_x_coeff[ 0]);
        fval("AMDX2   ", h->amd_x_coeff[ 1]);
        fval("AMDX3   ", h->amd_x_coeff[ 2]);
        fval("AMDX4   ", h->amd_x_coeff[ 3]);
        fval("AMDX5   ", h->amd_x_coeff[ 4]);
        fval("AMDX6   ", h->amd_x_coeff[ 5]);
        fval("AMDX7   ", h->amd_x_coeff[ 6]);
        fval("AMDX8   ", h->amd_x_coeff[ 7]);
        fval("AMDX9   ", h->amd_x_coeff[ 8]);
        fval("AMDX10  ", h->amd_x_coeff[ 9]);
        fval("AMDX11  ", h->amd_x_coeff[10]);
        fval("AMDX12  ", h->amd_x_coeff[11]);
        fval("AMDX13  ", h->amd_x_coeff[12]);
        fval("AMDX14  ", h->amd_x_coeff[13]);
        fval("AMDX15  ", h->amd_x_coeff[14]);
        fval("AMDX16  ", h->amd_x_coeff[15]);
        fval("AMDX17  ", h->amd_x_coeff[16]);
        fval("AMDX18  ", h->amd_x_coeff[17]);
        fval("AMDX19  ", h->amd_x_coeff[18]);
        fval("AMDX20  ", h->amd_x_coeff[19]);
        fval("AMDY1   ", h->amd_y_coeff[ 0]);
        fval("AMDY2   ", h->amd_y_coeff[ 1]);
        fval("AMDY3   ", h->amd_y_coeff[ 2]);
        fval("AMDY4   ", h->amd_y_coeff[ 3]);
        fval("AMDY5   ", h->amd_y_coeff[ 4]);
        fval("AMDY6   ", h->amd_y_coeff[ 5]);
        fval("AMDY7   ", h->amd_y_coeff[ 6]);
        fval("AMDY8   ", h->amd_y_coeff[ 7]);
        fval("AMDY9   ", h->amd_y_coeff[ 8]);
        fval("AMDY10  ", h->amd_y_coeff[ 9]);
        fval("AMDY11  ", h->amd_y_coeff[10]);
        fval("AMDY12  ", h->amd_y_coeff[11]);
        fval("AMDY13  ", h->amd_y_coeff[12]);
        fval("AMDY14  ", h->amd_y_coeff[13]);
        fval("AMDY15  ", h->amd_y_coeff[14]);
        fval("AMDY16  ", h->amd_y_coeff[15]);
        fval("AMDY17  ", h->amd_y_coeff[16]);
        fval("AMDY18  ", h->amd_y_coeff[17]);
        fval("AMDY19  ", h->amd_y_coeff[18]);
        fval("AMDY20  ", h->amd_y_coeff[19]);
		
		fval("AMDREX1 ", h->amdre_x_coeff[ 0]);
        fval("AMDREX2 ", h->amdre_x_coeff[ 1]);
        fval("AMDREX3 ", h->amdre_x_coeff[ 2]);
        fval("AMDREX4 ", h->amdre_x_coeff[ 3]);
        fval("AMDREX5 ", h->amdre_x_coeff[ 4]);
        fval("AMDREX6 ", h->amdre_x_coeff[ 5]);
        fval("AMDREX7 ", h->amdre_x_coeff[ 6]);
        fval("AMDREX8 ", h->amdre_x_coeff[ 7]);
        fval("AMDREX9 ", h->amdre_x_coeff[ 8]);
        fval("AMDREX10", h->amdre_x_coeff[ 9]);
        fval("AMDREX11", h->amdre_x_coeff[10]);
        fval("AMDREX12", h->amdre_x_coeff[11]);
        fval("AMDREX13", h->amdre_x_coeff[12]);
        fval("AMDREX14", h->amdre_x_coeff[13]);
        fval("AMDREX15", h->amdre_x_coeff[14]);
        fval("AMDREX16", h->amdre_x_coeff[15]);
        fval("AMDREX17", h->amdre_x_coeff[16]);
        fval("AMDREX18", h->amdre_x_coeff[17]);
        fval("AMDREX19", h->amdre_x_coeff[18]);
        fval("AMDREX20", h->amdre_x_coeff[19]);
        fval("AMDREY1 ", h->amdre_y_coeff[ 0]);
        fval("AMDREY2 ", h->amdre_y_coeff[ 1]);
        fval("AMDREY3 ", h->amdre_y_coeff[ 2]);
        fval("AMDREY4 ", h->amdre_y_coeff[ 3]);
        fval("AMDREY5 ", h->amdre_y_coeff[ 4]);
        fval("AMDREY6 ", h->amdre_y_coeff[ 5]);
        fval("AMDREY7 ", h->amdre_y_coeff[ 6]);
        fval("AMDREY8 ", h->amdre_y_coeff[ 7]);
        fval("AMDREY9 ", h->amdre_y_coeff[ 8]);
        fval("AMDREY10", h->amdre_y_coeff[ 9]);
        fval("AMDREY11", h->amdre_y_coeff[10]);
        fval("AMDREY12", h->amdre_y_coeff[11]);
        fval("AMDREY13", h->amdre_y_coeff[12]);
        fval("AMDREY14", h->amdre_y_coeff[13]);
        fval("AMDREY15", h->amdre_y_coeff[14]);
        fval("AMDREY16", h->amdre_y_coeff[15]);
        fval("AMDREY17", h->amdre_y_coeff[16]);
        fval("AMDREY18", h->amdre_y_coeff[17]);
        fval("AMDREY19", h->amdre_y_coeff[18]);
        fval("AMDREY20", h->amdre_y_coeff[19]);
		fval("REFBETA ", h->ref_Beta);
		fval("REFBETAP", h->ref_BetaP);
		fval("REFK1   ", h->ref_k1);
		fval("REFK2   ", h->ref_k2);
		cvalt("ASTRMASK", h->mask);

/*
 * handle copyright statements
 */
        if (strcmp(keyword, "NBFCPYRT") == 0)
        {
           totalNumBriefCopyRightStatements = atoi(value);
           h->numBriefCopyRightStatements = 0;
           if (h->briefCopyRightStatements != NULL)

/*****************
* readhdr page 4 *
*****************/

              free(h->briefCopyRightStatements);
        }
        if ((strncmp(keyword, "BFCPYRT", 7) == 0) &&
            (isalnum(keyword[7])                )   )
        {
           if (h->numBriefCopyRightStatements    <
               totalNumBriefCopyRightStatements   )
           {
              if (h->numBriefCopyRightStatements == 0)
                 h->briefCopyRightStatements = (char **)malloc(
                       totalNumBriefCopyRightStatements*sizeof(char*));

              strncpy(fullCharValue, &line[10], 70);
              fullCharValue[70] = '\0';
              fullCharValueLength = strlen(strtrim(remquotes(fullCharValue)));
              currentCopyRightStatement = h->numBriefCopyRightStatements;
              h->briefCopyRightStatements[currentCopyRightStatement] =
                    (char *)malloc(fullCharValueLength + 1);
              strncpy(h->briefCopyRightStatements[currentCopyRightStatement],
                      strtrim(&(fullCharValue[1])),
                      fullCharValueLength + 1);
              h->numBriefCopyRightStatements += 1;
           }
        }

        if (strcmp(keyword,"END     ")==0) break;
    }
    /*
     *  Decide if GSSS scan or special plate
     */
    if (strcmp(h->plate_id,"    ") == 0) {
        h->special_plate_flag = 'T';
    } else {
        h->special_plate_flag = 'F';
    }
    /*
     *  Fudge pixel size for early scan data
     *  (I'd be surprised if this works when roundoff error is included)
     */
    if ((h->special_plate_flag == 'F') && (h->x_pixel_size == 25.284)) {
         h->x_pixel_size = 25.28445;
         h->y_pixel_size = 25.28445;
    }
    /*
     *  See if AMD calibration exists
     */
    if (dtotal(h->amd_x_coeff,20) == 0.0) {
        h->amd_flag = 'F';
    } else {
        h->amd_flag = 'T';
    }
    /*
     *  See if AMDRE calibration exists
     */
    if (dtotal(h->amdre_x_coeff,20) == 0.0) {
        h->amdre_flag = 'F';
    } else {
        h->amdre_flag = 'T';
    }

    /*
     *  See if PPO calibration exists
     */
    if (dtotal(h->ppo_coeff,6) == 0.0) {
        h->ppo_flag = 'F';
    } else {
        h->ppo_flag = 'T';
    }

/*****************
* readhdr page 5 *
*****************/
    /*
     *  See if plate data exists
     */
    if (strcmp(h->plate_id,"    ") != 0) {
        h->plate_data_flag = 'T';
    } else {
        h->plate_data_flag = 'F';
    }
    /*
     *  Convert plate center position to radians
     */
    condmsrd(h->ra_h, h->ra_m, h->ra_s,
        h->dec_sign, h->dec_d, h->dec_m, h->dec_s,
        &h->plt_center_ra, &h->plt_center_dec);

} /* readhdr */
