/******************************************************************************
* C Source: Plate                                                             
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    This set of routines manipulate a structure (see plate.h) which contains 
*    data about a single plate, namely, it's id, region, center right         
*    ascension and declination in radians, epoch, and number of it's CD_ROM.  
*                                                                             
* Module List:                                                                
*    Compare_Plates                                                           
*       Lexigraphically compares the plates ids and volume labels of two      
*       plates.                                                               
*    Create_Empty_Plate                                                
*       Create memory for a plate and initialize fields to "zero".             
*    Delete_Plate                                                             
*       Free memory for a plate.                                              
*    Plate_Directory_Exists:                                                  
*       Checks if the directory for a plate's image data exists.              
*    Read_Version_1_Plate_Info:
*       Read data for a single Plate from version 1 of plate list file.
*    Read_Version_2_Plate_Info:
*       Read data for a single Plate from version 2 of plate list file.
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 14-NOV-94 by J. Doggett,                                    
*       Updated Read_Plate_Info function.                                     
*    Redelivered: 30-NOV-94 by J. Doggett,                                    
*       GetImage Version 1.1.                                                 
*                                                                             
*    Updated    : 30-APR-95 by J. Doggett,                                    
*       Updated Create_Plate and Read_Plate_Info functions.                   
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
*                                                                             
*    Updated    : 05-JAN-96 by J. Doggett,                                    
*       Updated Read_Plate_Info function.                                     
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
*                                                                             
*    Updated    : 12-SEP-96 by J. Doggett,                                    
*       Updated Create_Plate and Read_Plate_Info functions.                   
*          functions.                                                         
*       Removed the now obsolete function Plates_Disc_Is_Available.           
*    Updated    : 21-NOV-96 by J. Doggett,                                    
*       Renamed original Read_Plate_Info function to Read_Version_1_Plate_Info
*          and also updated it.
*       Created new Read_Version_2_Plate_Info function.
*    Updated    : 22-NOV-96 by J. Doggett,
*       Renamed Create_Plate to Create_Empty_Plate because of change in
*       functionality.
*    Updated    : 17-NOV-96 by J. Doggett,
*       Redelivered Read_Version_1_Plate_Info
*       Delivered Read_Version_2_Plate_Info
*******************************************************************************

/* *** Plate Introduction ***                                              2 */


/******************************************************************************
*    Updated    : 19-FEB-97 by J. Doggett,                                    
*       Redelivered Compare_Plates functions.				      
*    Redelivered: xx-xxx-xx by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
*                                                                             
* Copyright (c) 1993-1997, Association of Universities for Research in        
* Astronomy, Inc.  All Rights Reserved.                                       
* Produced under National Aeronautics and Space Administration grant          
* NAG W-2166.                                                                 
******************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* access function not prototyped in standard place */
#if defined __VMS
#  include <unixio.h>
#elif defined __UNIX
#  include <unixstd.h>
#else
#endif

#include "plate.h"

#include "astronmy.h"
#include "filenams.h"
#include "generic.h"

/******************************************************************************
* Module:                                                                     
*    Compare_Plates                                                           
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Lexigraphically compares the plates ids and volume labels of two plates. 
*    If either of the volume labels is zero length, then the volume labels    
*    are not compared.							      
*                                                                             
* Return Value:                                                               
*    int: The result from using strcmp to compare the two plate ids and       
*         volume labels                                                       
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Plate_1                                                               
*          PLATE, Reference                                                   
*             The first plate to compare.                                     
*       Plate_2                                                               
*          PLATE, Reference                                                   
*             The second plate to compare.                                    
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 24-OCT-96 by J. Doggett,                                    
*       Modified to compare volume labels in addition to plate ids.           
*    Redelivered: 19-FEB-97 by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
******************************************************************************/
/******************************
* Compare_Plates declarations *
******************************/
int   Compare_Plates( Plate_1, Plate_2 )
PLATE Plate_1;
PLATE Plate_2;
{
   int compareResult;
/***************************************
* Compare_Plates begin executable code *
***************************************/
   if (Plate_1->Id == Plate_2->Id)
      compareResult = 0;
   else
      compareResult = strcmp( Plate_1->Id, Plate_2->Id );

   if (compareResult == 0)
      if ((strlen(Plate_1->volumeLabel) != 0) &&
	  (strlen(Plate_2->volumeLabel) != 0)   )
      {
									      
/* *** Compare_Plates ***                                                  2 */


	 if (Plate_1->volumeLabel == Plate_2->volumeLabel)
	    compareResult = 0;
	 else
            compareResult = 
	       strcmp( Plate_1->volumeLabel, Plate_2->volumeLabel );
      }
   return compareResult;
}
/*************************************
* Compare_Plates end executable code *
*************************************/

/******************************************************************************
* Module:                                                                     
*    Create_Empty_Plate                                                       
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Allocates memory to hold data for a plate, and inserts the data into the 
*    memory.                                                                  
*                                                                             
* Return Value:                                                               
*    PLATE: The newly created plate.                                          
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       none
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 30-APR-95 by J. Doggett,                                    
*       Updated to handle new volumeLabel field for plate structure.          
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
*                                                                             
*    Updated    : 22-NOV-96 by J. Doggett,                                    
*       Renamed to Create_Empty_Plate because functionality changed to only
*       allocate memory and initialize fields to "zero".
*    Redelivered: 22-NOV-96 by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
******************************************************************************/
/**********************************
* Create_Empty_Plate declarations *
**********************************/
PLATE  Create_Empty_Plate()

{
   PLATE Plate; /* Newly created plate structure         */

/*******************************************
* Create_Empty_Plate begin executable code *
*******************************************/
   
   Plate = ( PLATE )Allocate_Memory( sizeof( *Plate ) );

   Plate->Id[0]                 = (char)NULL;
   Plate->Region[0]             = (char)NULL;
   Plate->Center_RA             = 0.        ;
   Plate->Center_Dec            = 0.        ;
   Plate->Epoch                 = 0.        ;
   Plate->volumeLabel[0]        = (char)NULL;
   Plate->digitizedSurvey[0]    = (char)NULL;

/* *** Create_Empty_Plate ***                                              2 */


   Plate->astronomicalSurvey[0] = (char)NULL;
   Plate->bandpass[0]           = (char)NULL;
   Plate->maskFile[0]           = (char)NULL;
   Plate->calibrationFile[0]    = (char)NULL;
   Plate->plateInfoVersion      = 0.        ;

   return Plate;
}
/*****************************************
* Create_Empty_Plate end executable code *
*****************************************/

/******************************************************************************
* Module:                                                                     
*    Delete_Plate                                                             
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Free memory previously allocated for a plate.                            
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Plate                                                                 
*          PLATE, Reference                                                   
*             The plate we're deleting from memory.                           
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/****************************
* Delete_Plate declarations *
****************************/
void  Delete_Plate( Plate )
PLATE Plate;
{
/*************************************
* Delete_Plate begin executable code *
*************************************/
   if ( Plate != NULL )
      free( Plate );

   return;
}
/***********************************
* Delete_Plate end executable code *
***********************************/

/******************************************************************************
* Module:                                                                     
*    Plate_Directory_Exists                                                   
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Checks if the directory for a plate's image data exists.                 
*                                                                             
* Return Value:                                                               
*    BOOL: TRUE if the directory exists, FALSE otherwise.                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Plate_Name                                                            
*          STRING, Reference                                                  
*             Name of plate for whose's directory we're looking.              
*       Image_Root_Directory                                                  
*          STRING, Reference                                                  
*             Name of parent directory for all plates' directories.           
*    Input/Output Arguments:                                                  
*          none                                                               
*    Output Arguments:                                                        
*          none                                                               
*                                                                             
* History:                                                                    
*    Created    : 07-SEP-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
/**************************************
* Plate_Directory_Exists declarations *
**************************************/
BOOL   Plate_Directory_Exists( Plate_Name, Image_Root_Directory )
STRING Plate_Name          ;
STRING Image_Root_Directory;
{
   STRING Temp_File_Name     ; /* Temp place for base directory file name */
   STRING Directory_File_Name; /* Full name for directory file            */
   FILE   *File              ; /* Pointer to file when we try to open it  */
   BOOL   Exists             ; /* Holds the return value                  */

/***********************************************
* Plate_Directory_Exists begin executable code *
***********************************************/
#if defined __VMS
   Temp_File_Name = ( STRING )Allocate_Memory( strlen( Plate_Name ) + 5 );
   strcpy( Temp_File_Name, Plate_Name );
   strcat( Temp_File_Name, ".DIR" );

#elif defined __UNIX
   Temp_File_Name = ( STRING )Allocate_Memory( strlen( Plate_Name ) + 1 );
   strcpy( Temp_File_Name, Plate_Name );

#else
   fprintf(stderr, "Unknown file system error in Plate_Diretory_Exits");
   exit(EXIT_FAILURE);


/* *** Plate_Directory_Exits ***                                           2 */


#endif

   Directory_File_Name = NULL;
   if (Set_FileNames_Case(Image_Root_Directory, Temp_File_Name)) {
      Directory_File_Name =
              Build_Full_File_Name( Image_Root_Directory, "", Temp_File_Name );

      if ( access( Directory_File_Name, 0 ) == 0 )
         Exists = TRUE;
      else
         Exists = FALSE;
   }
   else
      Exists = FALSE;
   fprintf(stderr, Image_Root_Directory);
   free( Temp_File_Name );
   if (Directory_File_Name != NULL)
      free( Directory_File_Name );

   return Exists;
}
/*********************************************
* Plate_Directory_Exists end executable code *
*********************************************/

/******************************************************************************
* Module:                                                                     
*    Read_Version_1_Plate_Info
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Read data for Plate from version 1 plate list file.
*                                                                             
* Return Value:                                                               
*    PLATE: The plate read from the file.                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       InputFile                                                             
*          FILE*, Value                                                       
*             File pointer to the input file.                                 
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 17-DEC-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 01-JUN-94 by J. Doggett,                                    
*       Modified for the change in format of the input file from records      
*       separated with a new-line character to records separated with a       
*       dollar sign ($).                                                      
*    Redelivered: 30-NOV-94 by J. Doggett,                                    
*       GetImage Version 1.1.                                                 
*                                                                             
*    Updated    : 30-APR-95 by J. Doggett,                                    
*       Now expects volume label rather than disc number.                     
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
*                                                                             
*    Updated    : 05-JAN-96 by J. Doggett,                                    
*       Modified to convert input strings to upper case.                      
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
*                                                                             
*    Updated    : 12-SEP-96 by J. Doggett,                                    
*       Modified to no longer put disc number into PLATE structure since disc 
*       number has become obsolete.                                           
*    Updated    : 21-NOV-96 by J. Doggett,                                    
*       Replace Comment_Character variable with PLATE_RECORD_COMMENT_CHARACTER
*          macro.
*       Removed definition of PLATE_RECORD_EOL_CHAR because not defined in
*          plate.h.
*       Removed putting source into Plate structure.
*    Updated    : 17-DEC-96 by J. Doggett,                                    
*       Removed error message on failure to read region.
*       Modified to clear input on failure to read region.
*    Redelivered: 21-NOV-96 by J. Doggett,                                    
******************************************************************************/

/* *** Read_Version_1_Plate_Info ***                                       2 */


/******************************************************************************
*       GetImage Version 2.0x.                                                
******************************************************************************/
/*****************************************
* Read_Version_1_Plate_Info declarations *
*****************************************/
PLATE Read_Version_1_Plate_Info( InputFile )
FILE*   InputFile;
{
   PLATE Plate; /* Place for plate data read in */

/* Variables for reading data */
   char   Region  [ 7 ]  ;
   char   Plate_Id[ 5 ]  ;
   int    RA_Hours       ; /* Plate center                       */
   int    RA_Minutes     ;
   double RA_Seconds     ;
   char   Dec_Sign       ;
   int    Dec_Degrees    ;
   int    Dec_ArcMinutes ;
   double Dec_ArcSeconds ;
   double RA_Radians     ;
   double Dec_Radians    ;
   double Epoch          ;
   char   Source[ 4 ]    ;
   char   volumeLabel[24];
   int    Temp_Char     ;

/**************************************************
* Read_Version_1_Plate_Info begin executable code *
**************************************************/
/* Read plate region, id */
   Region[0] = PLATE_RECORD_COMMENT_CHARACTER;
   while ( Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
   {
      Region[0] = getc(InputFile);
      if (Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
                  ) ;
      else
      {
         ungetc(Region[0], InputFile);
         switch(fscanf(InputFile, "%s", Region))
         {
            case 1:
               break;
            case EOF:
               return ( PLATE )EOF;
            default:
               while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                               &&
                   (Temp_Char                    != EOF )
                  ) ;
               return (PLATE)NULL;
         }
      }
   }
   Convert_String_to_Upper_Case(Region);


/* *** Read_Version_1_Plate_Info ***                                       3 */


   switch(fscanf(InputFile, "%s", Plate_Id))
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }
   Convert_String_to_Upper_Case(Plate_Id);

/* Read plate position */
   switch( Read_DMS( InputFile, &RA_Radians, &Dec_Radians ) )
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }

/* Read epoch, source, volume label */
   switch( fscanf( InputFile, "%lf %s %s",
                   &Epoch, Source, volumeLabel ) )
   {
      case 3  :
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
   
         Convert_String_to_Upper_Case(volumeLabel);
   
         Plate = Create_Empty_Plate();
   
         Plate->plateInfoVersion = 1.            ;
         strcpy( Plate->Id, Plate_Id )           ;
         strcpy( Plate->Region, Region )         ;
         Plate->Center_RA  = RA_Radians          ;
         Plate->Center_Dec = Dec_Radians         ;
         Plate->Epoch      = Epoch               ;
         strcpy( Plate->volumeLabel, volumeLabel);
   
         return Plate;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;

/* *** Read_Version_1_Plate_Info ***                                       4 */


   }
}
/************************************************
* Read_Version_1_Plate_Info end executable code *
************************************************/

/******************************************************************************
* Module:                                                                     
*    Read_Version_2_Plate_Info
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Read data for Plate from version 2 plate list file.
*                                                                             
* Return Value:                                                               
*    PLATE: The plate read from the file.                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       InputFile                                                             
*          FILE*, Value                                                       
*             File pointer to the input file.                                 
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 21-NOV-96 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 17-DEC-96 by J. Doggett,  
*       GetImage Version 2.0x.              
******************************************************************************/
/*****************************************
* Read_Version_2_Plate_Info declarations *
*****************************************/
PLATE Read_Version_2_Plate_Info( InputFile )
FILE*   InputFile;
{
   PLATE Plate; /* Place for plate data read in */

/* Variables for reading data */
   char   Region  [ 7 ]  ;
   char   Plate_Id[ 5 ]  ;
   int    RA_Hours       ; /* Plate center                       */
   int    RA_Minutes     ;
   double RA_Seconds     ;
   char   Dec_Sign       ;
   int    Dec_Degrees    ;
   int    Dec_ArcMinutes ;
   double Dec_ArcSeconds ;
   double RA_Radians     ;
   double Dec_Radians    ;
   double Epoch          ;
   char   volumeLabel[24];
   int    Temp_Char     ;
   char   astronomicalSurvey[51];
   char   bandpass[16];
   char   calibrationFile[81];

/**************************************************
* Read_Version_2_Plate_Info begin executable code *
**************************************************/
/* Read plate region, id */

/* *** Read_Version_2_Plate_Info ***                                       2 */


   Region[0] = PLATE_RECORD_COMMENT_CHARACTER;
   while ( Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
   {
      Region[0] = getc(InputFile);
      if (Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
                  ) ;
      else
      {
         ungetc(Region[0], InputFile);
         switch(fscanf(InputFile, "%s", Region))
         {
            case 1:
               break;
            case EOF:
               return ( PLATE )EOF;
            default:
               while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                               &&
                   (Temp_Char                    != EOF )
                  ) ;
               return (PLATE)NULL;
         }
      }
   }
   Convert_String_to_Upper_Case(Region);

   switch(fscanf(InputFile, "%s", Plate_Id))
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }
   Convert_String_to_Upper_Case(Plate_Id);

/* Read plate position */
   switch( Read_DMS( InputFile, &RA_Radians, &Dec_Radians ) )
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }

/* Read epoch, volume label */
   switch( fscanf( InputFile, "%lf %s %s %s %s",

/* *** Read_Version_2_Plate_Info ***                                       3 */


                   &Epoch,
		   volumeLabel, 
		   astronomicalSurvey,
		   bandpass,
		   calibrationFile) )
		   
   {
      case 5  :
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
   
         Convert_String_to_Upper_Case(volumeLabel);
   
         Plate = Create_Empty_Plate();
   
         Plate->plateInfoVersion = 2.            ;
         strcpy( Plate->Id, Plate_Id )           ;
         strcpy( Plate->Region, Region )         ;
         Plate->Center_RA  = RA_Radians          ;
         Plate->Center_Dec = Dec_Radians         ;
         Plate->Epoch      = Epoch               ;
         strcpy( Plate->volumeLabel, volumeLabel);
         strncpy(Plate->digitizedSurvey, &(volumeLabel[14]), 4);
         strcpy(Plate->astronomicalSurvey, astronomicalSurvey);
         strcpy(Plate->bandpass, bandpass);
         strcpy(Plate->calibrationFile, calibrationFile);
		 /*printf("%s\n",calibrationFile);*/
      
         return Plate;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
      }
}
/************************************************
* Read_Version_2_Plate_Info end executable code *
************************************************/

/******************************************************************************
* Module:                                                                     
*    Read_Version_3_Plate_Info
*                                                                             
* File:                                                                       
*    plate.c                                                                  
*                                                                             
* Description:                                                                
*    Read data for Plate from version 3 plate list file.
*                                                                             
* Return Value:                                                               
*    PLATE: The plate read from the file.                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       InputFile                                                             
*          FILE*, Value                                                       
*             File pointer to the input file.                                 
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 26-JUL-2005 by Brian McLean,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                                                                
*       GetImage Version 3.x.              
******************************************************************************/
/*****************************************
* Read_Version_3_Plate_Info declarations *
*****************************************/
PLATE Read_Version_3_Plate_Info( InputFile )
FILE*   InputFile;
{
   PLATE Plate; /* Place for plate data read in */

/* Variables for reading data */
   char   Region  [ 7 ]  ;
   char   Plate_Id[ 5 ]  ;
   int    RA_Hours       ; /* Plate center                       */
   int    RA_Minutes     ;
   double RA_Seconds     ;
   char   Dec_Sign       ;
   int    Dec_Degrees    ;
   int    Dec_ArcMinutes ;
   double Dec_ArcSeconds ;
   double RA_Radians     ;
   double Dec_Radians    ;
   double Epoch          ;
   char   volumeLabel[24];
   int    Temp_Char     ;
   char   astronomicalSurvey[51];
   char   bandpass[16];
   char   calibrationFile[81];
   char   maskFile[81];

/**************************************************
* Read_Version_3_Plate_Info begin executable code *
**************************************************/
/* Read plate region, id */
   Region[0] = PLATE_RECORD_COMMENT_CHARACTER;
   while ( Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
   {
      Region[0] = getc(InputFile);
      if (Region[0] == PLATE_RECORD_COMMENT_CHARACTER)
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
                  ) ;
      else
      {
         ungetc(Region[0], InputFile);
         switch(fscanf(InputFile, "%s", Region))
         {
            case 1:
               break;
            case EOF:
               return ( PLATE )EOF;
            default:
               while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                               &&
                   (Temp_Char                    != EOF )
                  ) ;
               return (PLATE)NULL;
         }
      }
   }
   Convert_String_to_Upper_Case(Region);

   switch(fscanf(InputFile, "%s", Plate_Id))
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }
   Convert_String_to_Upper_Case(Plate_Id);

/* Read plate position */
   switch( Read_DMS( InputFile, &RA_Radians, &Dec_Radians ) )
   {
      case 1:
         break;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
   }
/* Read epoch, volume label */
   switch( fscanf( InputFile, "%lf %s %s %s %s %s",
		   &Epoch,
		   volumeLabel, 
		   astronomicalSurvey,
		   bandpass,
		   maskFile,
		   calibrationFile) )
	   
   {
      case 6  :
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
   
         Convert_String_to_Upper_Case(volumeLabel);
   
         Plate = Create_Empty_Plate();
   
         Plate->plateInfoVersion = 3.            ;
         strcpy( Plate->Id, Plate_Id )           ;
         strcpy( Plate->Region, Region )         ;
         Plate->Center_RA  = RA_Radians          ;
         Plate->Center_Dec = Dec_Radians         ;
         Plate->Epoch      = Epoch               ;
         strcpy( Plate->volumeLabel, volumeLabel);
         strncpy(Plate->digitizedSurvey, &(volumeLabel[14]), 4);
         strcpy(Plate->astronomicalSurvey, astronomicalSurvey);
         strcpy(Plate->bandpass, bandpass);
		 strcpy(Plate->maskFile, maskFile);
         strcpy(Plate->calibrationFile, calibrationFile);      
         return Plate;
      case EOF:
         return ( PLATE )EOF;
      default:
         while (( (Temp_Char=getc(InputFile)) != PLATE_RECORD_EOL_CHAR)
                            &&
                (Temp_Char                    != EOF )
               ) ;
         return ( PLATE )NULL;
      }
}
/************************************************
* Read_Version_3_Plate_Info end executable code *
************************************************/

