/*
 * This program was developed for use with the NASA's Digitial Sky Survey to
 * deal with the Solaris 2.x volume manager, and handle the case where the
 * volume manager is not running or not managing the CDROM device.  It is
 * in a seperate program to limit the security exposures when the program
 * is installed setuid to root to deal with the case where the CDROM drive
 * is not being managed by vold.  It would be too complex to analyze the
 * implications of installing getimage setuid to root.  As this program
 * is linked by default, you have the possibility of the application picking
 * up the wrong versions of the c or volmgt shared libraries, or possibly
 * someone replacing these libraries.  If this is a concern the application
 * should be statically linked.  If you have volume management as under
 * Solaris 2.2 or later define -DVOLMGT
 *
 * If you are not using volume management to manage the device, the volume
 * manager is not running, or you are running Solaris 1.x, the device name
 * must be of the form /dev/sr?.  On Solaris 2.x systems, this implies that
 * if you are not running the volume manager or not managing the CD-ROM
 * devices that you either have the BSD compatibility package loaded which
 * makes the links to the CDROM devices from /dev.  I am not sure that ucblinks
 * is smart enough to check whether target 6 is really a CDROM, or look for
 * other targets.
 *
 * The $DSS_DIR should be replaced with what ever you plan to set DSS_DIR
 * to in the getimage shell script.  It should also work if it is in another
 * location in the user's path.
 *
 * To comple under Solaris 2.x and install:
 *
 * %cc -O -DVOLMGT ss_mount.c -o ss_mount -lvolmgt
 * %cp ss_mount $DSS_DIR                (where ever getimage lives)
 * %chown root $DSS_DIR/ss_mount        (only necessary if not using vold)
 * %chmod 4711 $DSS_DIR/ss_mount        (only necessary if not using vold)
 *
 * To compile under Solaris 1.x and install:
 *
 * %cc -O ss_mount.c -o ss_mount
 * %cp ss_mount $DSS_DIR                (where ever getimage lives)
 * %chown root $DSS_DIR/ss_mount
 * %chmod 4711 $DSS_DIR/ss_mount
 *
 * If you have security concerns with installing this program setuid as
 * root, or think you have found an unexpected loophole or problem, send
 * your questions, comments, or problems to:
 *
 *                                      getimage@stsci.edu
 *
 * Replaced non-ANSI standard function strdup with malloc and strcpy.
 *                      J. Doggett, 01-JUN-1995
 */

/******************************************************************************
* Copyright (c) 1994, 1995 Association of Universities for Research in        *
* Astronomy, Inc.  All Rights Reserved.                                       *
* Produced under National Aeronautics and Space Administration grant          *
* NAG W-2166.                                                                 *
******************************************************************************/

#include <stdio.h>
#ifdef VOLMGT
#include <volmgt.h>
#include <sys/mnttab.h>
#endif

#include <string.h>


#include <sys/param.h>

#ifdef __osf__
#include <sys/types.h>
char *realpath(const char *, char *);
#define __STDC__
#endif

#ifdef __STDC__
#include <unistd.h>
#include <stdlib.h>
#else
char *malloc(), *realloc(), *getcwd(), *getenv();
#endif

#include <dirent.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>

/* #include <alloca.h> */ /* Replaced use of alloca with malloc */
#ifdef __sgi
#ifndef __STDC__
#define __STDC__
#endif
#else
/* char *alloca(int); *//* Replaced use of alloca with malloc */
#endif

int olduid = -1;
int verbose = 0;


/*
 * This is an array of recognized devices, you may have to change the code
 * to have the system recognize your devices.  You should modify this array
 * so that it only knows about the devices you might have.
 */
static char *allowed_devices[] = {
#if defined(sun) || defined(__sun)
"/dev/sr0",
"/dev/sr1",
/*
 * Other possible Solaris 1.x or 2.x names
 */
#ifdef notdef
"/dev/sr2",
"/dev/sr3",
"/dev/sr4",
"/dev/sr5",
"/dev/sr6",
"/dev/sr7",
"/dev/sr8",
"/dev/sr9",
#endif
/*
 * Possible Solaris 2.x names
 */
#ifdef _SVR4
"/dev/dsk/c0t6d0s2",
"/dev/dsk/c1t6d0s2",
#ifdef notdef
"/dev/dsk/c0t5d0s2",
"/dev/dsk/c1t5d0s2",
#endif
#endif
#endif
/*
 * Possible sgi devices
 */
#ifdef __sgi
"/dev/scsi/c0d4l0",
"/dev/scsi/sc0d4l0",
"/dev/scsi/c1d4l0",
"/dev/scsi/sc1d4l0",
#endif
/*
 * Possible osf devices
 */
#ifdef __osf__
"/dev/rz3c",
"/dev/rz4c",
#endif
NULL,
};

#ifdef VOLMGT
static char **volmgt_devices;
static char **volmgt_dirs;
#endif
#ifdef __sgi
char *sgidev = 0;
char *sgidir = 0;
#endif

static int volmgt_devcount = 0;

static int do_mount(
#ifdef __STDC__
char *, char **, int
#endif
);

static int do_unmount(
#ifdef __STDC__
char *, char **, int
#endif
);

/*
 * The following checks the device name against the array above to
 * see if it is an allowed device.
 */

static int check_device(
#ifdef __STDC__
char *dev)
#else
dev)
char *dev;
#endif
{
    int j;

    for(j=0;;j++) {
        if(allowed_devices[j] == NULL) {
            fprintf(stderr,
                    "%s is not a recognized CD-ROM device\n", dev);
            return(0);
        }
        if(strcmp(dev, allowed_devices[j]) == 0)
            break;
    }
    return(1);
}


main(
#ifndef __STDC__
argc, argv)
int argc;
char **argv;
#else
int argc, char **argv)
#endif
{

/*
 * Take away effective root privs and make the effective uid his own for
 * file access checks and the like.  This assumes the program was installed
 * setuid or is being run by root.  As long as either the effective or real
 * uid's is root, it should work.
 */

    if(geteuid() == 0 || getuid() == 0) {
        olduid = getuid();
        setuid(0);
        seteuid(olduid);
    }

    if(argc < 2) {
        fprintf(stderr,
         "Usage: %s mount [-v] [device dir label]\nUsage: %s unmount device\n",
         *argv, *argv);
        exit(1);
    }

    if(!sol_init_mount()) {
        if(olduid < 0) {
#ifdef VOLMGT
fprintf(stderr,
  "Sorry, your cdrom cannot be mounted.  Either you don't have vold running\n");
fprintf(stderr,
  "or vold is not managing your CDROM devices.  This program was also not\n");
fprintf(stderr,
  "installed to run setuid to root, so it can't get round the problem\n");
#else
#ifdef __sgi
fprintf(stderr,
  "Sorry, your cdrom cannot be mounted.  cdromd is not configured to\n");
fprintf(stderr,
  "manage the drive and this program was not installed setuid to root\n");
#else
fprintf(stderr,
 "Sorry, your cdrom cannot be mounted.  %s must be installed setuid to root\n",
  *argv);
#endif
#endif
            exit(1);
        }
    }

    if(strcasecmp(argv[1], "mount") == 0) {
        if(strcasecmp(argv[2], "-v") == 0) {
            verbose = 1;
            exit(do_mount(*argv, argv+3, argc-3));
        }
        exit(do_mount(*argv, argv+2, argc-2));
    }
    if(strcasecmp(argv[1], "unmount") == 0)
        exit(do_unmount(*argv, argv+2, argc-2));
    if(strcasecmp(argv[1], "umount") == 0)
        exit(do_unmount(*argv, argv+2, argc-2));
    fprintf(stderr, "%s: Unknown command: %s\n", *argv, argv[1]);
    exit(1);
}

int sol_init_mount(
#ifdef __STDC__
void
#endif
)
{
#ifndef VOLMGT
#ifdef __sgi
    FILE *fp;
    char dev[256], dir[256], fstype[32], buf[1024];
    fp = fopen("/etc/fsd.tab", "r");
    if(!fp)
        return(0);
    while(fgets(buf, sizeof(buf), fp)) {
        sscanf(buf, "%s%s%s", dev, dir, fstype);
        if(strcmp(fstype, "cdrom") == 0 || strcmp(fstype, "iso9660") == 0) {
            if(sgidev) {
                fprintf(stderr, "This program only supports one cdrom drive ");
                fprintf(stderr, "on SGI systems.\n Please use the drive ");
                fprintf(stderr, "associated with the device %s\n", sgidev);
                fclose(fp);
                return(1);
            }
/*          sgidev = strdup(dev); */ /* strdup is non-ANSI standard */
            sgidev = (char *)malloc(strlen(dev) + 1);
            strcpy(sgidev, dev);

/*          sgidir = strdup(dir); */ /* strdup is non-ANSI standard */
            sgidir = (char *)malloc(strlen(dir) + 1);
            strcpy(sgidir, dir);

            volmgt_devcount = 1;
        }
    }
    fclose(fp);
    return(volmgt_devcount);
#else
    return(0);
#endif
#else
    const char *root;
    char *newdir, *newdev, *newdirs;
    DIR *dp;
    struct dirent *de;

    if(!volmgt_running())               /* Is volume management running */
        return(0);

    root = volmgt_root();

    if(!root)                           /* Get the root         */
        return(0);

/*  newdir = alloca(strlen(root) + 9);*/
    newdir = malloc(strlen(root) + 9);
    if(!newdir)
        return(0);
    sprintf(newdir, "%s/dev/dsk", root);

    dp = opendir(newdir);               /* Look for mount points */

    if(!dp){
       free(newdir); /* Added for change of alloca to malloc */
        return(0);
    }

/*
 * The following scans the directory and convers the directory names to
 * device names and then asked the volume manager if it is managing that
 * volume.  The last part is just an extra safe guard.  If the volume
 * manager is managing the volume, we add the directory to our list.  We
 * do things this we since we can't trust /cdrom to be the place to find
 * the files or trust it to maintain /cdrom correctly.  We can also manage
 * multiple cdrom drives this way.
 */
    while((de = readdir(dp)) != NULL) {
        if(strcmp(de->d_name, ".") == 0 ||
           strcmp(de->d_name, "..") == 0)
                continue;
/*      newdev = alloca(strlen(de->d_name) + strlen("/dev/dsk/d0s2") + 1);*/
        newdev = malloc(strlen(de->d_name) + strlen("/dev/dsk/d0s2") + 1);
        if(!newdev){
	   free(newdir); /* Added for change of alloca to malloc */
            return(0);
	 }
        sprintf(newdev, "/dev/dsk/%sd0s2", de->d_name);
        if(!volmgt_inuse(newdev)) {
            newdev[strlen(newdev)-1] = '0';
            if(!volmgt_inuse(newdev))
                continue;
        }
/*          newdirs = alloca(strlen(newdir) + strlen(de->d_name) + 2);*/
            newdirs = malloc(strlen(newdir) + strlen(de->d_name) + 2);
            if(!newdirs){
	       free(newdir); /* Added for change of alloca to malloc */
	       free(newdev); /* Added for change of alloca to malloc */
                return(0);
	    }
            sprintf(newdirs, "%s/%s", newdir, de->d_name);
            add_to_volmgt_array(newdev, newdirs);
    }
    closedir(dp);

    free(newdir ); /* Added for change of alloca to malloc */
    free(newdev ); /* Added for change of alloca to malloc */
    free(newdirs); /* Added for change of alloca to malloc */
    return(volmgt_devcount);
#endif
}

#ifdef VOLMGT
static int add_to_volmgt_array(
#ifdef __STDC__
char *newdev, char *newdir)
#else
newdev, newdir)
char *newdev, *newdir;
#endif
{
    char *sp, *sp2;

/*  sp = strdup(newdev); */ /* strdup is non-ANSI standard */
    sp = (char *)malloc(strlen(newdev) + 1);
    strcpy(sp, newdev);

    if(!sp)
        return(0);

/*  sp2 = strdup(newdir);*/ /* strdup is non-ANSI standard */
    sp2 = (char *)malloc(strlen(newdir) + 1);
    strcpy(sp2, newdir);

    if(!sp2) {
        free(sp);
        return(0);
    }

    if(!volmgt_devices) {
        volmgt_devices = (char **)malloc(sizeof(char *));
        volmgt_dirs = (char **)malloc(sizeof(char *));
        if(volmgt_devices && volmgt_dirs)
            volmgt_devcount++;
        else {
            free(sp);
            free(sp2);
            if(volmgt_devices)
                free(volmgt_devices);
            return(0);
        }
    }

    else {
        volmgt_devcount++;
        volmgt_devices = (char **) realloc(volmgt_devices, sizeof(char *) *
                                           volmgt_devcount);
        volmgt_dirs = (char **) realloc(volmgt_dirs, sizeof(char *) *
                                        volmgt_devcount);
        if(!volmgt_devices || !volmgt_dirs) {
            volmgt_devcount = 0;
            free(sp);
            free(sp2);
            if(volmgt_devices)
                free(volmgt_devices);
            return(0);
        }
    }

    volmgt_devices[volmgt_devcount-1] = sp;
    volmgt_dirs[volmgt_devcount-1] = sp2;
    return(1);

}
#endif

static int do_mount(
#ifdef __STDC__
char *prog, char **argv, int argc)
#else
prog, argv, argc)
char *prog, **argv;
int argc;
#endif
{
#ifdef VOLMGT
    DIR *dp;
    struct dirent *de;
    int num_found;
#endif
#ifdef __sgi
#endif

    struct stat stbuf;
    char dir[MAXPATHLEN];
    char *mount_prog, *flag, *fstype;
    char *nargv[7], *tdir, *sp, *nsp;
    int exit_code, status, i, j;
    pid_t pid;

    if(argc == 0) {     /* Mount list */
#ifdef VOLMGT
        if(volmgt_devcount == 0) {
            fprintf(stderr, "%s: No CD-ROM's managed by vold\n", prog);
            return(1);
        }
        for(i=0;i!=volmgt_devcount;i++) {
            dp = opendir(volmgt_dirs[i]);
            if(dp) {
                while((de = readdir(dp)) != NULL) {
                    if(strcmp(de->d_name, ".") == 0 || strcmp(de->d_name, "..")
                       == 0)
                        continue;
                    printf("%s\n", de->d_name);
                }
            }
            else
                closedir(dp);
        }
#else
        fprintf(stderr,
                "%s: Mount list is not supported without volume management\n",
                prog);
        return(1);
#endif
    }
/*
 * The following watches for the cdrom we want to get mounted.  If it does
 * not appear, we return an error.  This should function with multiple
 * CDROM readers.
 */
    if(volmgt_devcount != 0 && argc == 3) {
#ifdef VOLMGT
/*
 * Tell volume management to look for a device
 */
        volmgt_check(NULL);
/*
 * Wait for a CDROM to show up online
 */
        for(j=0;j!=20;j++) {
            num_found = 0;
            for(i=0;i!=volmgt_devcount;i++) {
                dp = opendir(volmgt_dirs[i]);
                if(dp) {
                    while((de = readdir(dp)) != NULL) {
                        if(strcmp(de->d_name, ".") == 0 ||
                           strcmp(de->d_name, "..") == 0)
                            continue;
                        num_found++;
/*
 * We have found a mounted device.  Now check to see if it is the
 * one we want.  Then translate it to a mount point.
 */
                        if(strcasecmp(de->d_name, argv[2]) == 0) {
                            FILE *fp;
                            struct mnttab mntpref, mntent;
                            char *mntdev;
                            closedir(dp);
                            fp = fopen("/etc/mnttab", "r");
                            if(!fp) {
                                perror("/etc/mnttab");
                                exit(1);
                            }
/*                          mntdev = alloca(strlen(volmgt_dirs[i]) +
                                            strlen(de->d_name) + 2);*/
                            mntdev = malloc(strlen(volmgt_dirs[i]) +
                                            strlen(de->d_name) + 2);
                            sprintf(mntdev, "%s/%s", volmgt_dirs[i],
                                    de->d_name);

                            memset(&mntpref, 0, sizeof(mntpref));
                            mntpref.mnt_special = mntdev;
                            status = getmntany(fp, &mntent, &mntpref);
                            fclose(fp);

/*
 * The device is out there, loaded, but not mounted yet.  We claim that it
 * wasn't found yet and make another pass
 */
                            if(status == -1) {
                                num_found--;
                                continue;
                            }

                            if(verbose) {
                                printf("%s\n", mntent.mnt_mountp);
                                printf("%s\n", de->d_name);
                            }
			    free(mntdev); /* Added for change of alloca to malloc */
                            return(0);
                        }
                    }
                    closedir(dp);
                }
            }
            if(num_found != volmgt_devcount)
                sleep(1);
            else
                break;
        }
        fprintf(stderr,"%s: Volume %s was not available\n", prog, argv[2]);
        return(1);
#endif
#ifdef __sgi
        for(j=0;j!=20;j++) {
                if(stat(sgidir, &stbuf) < 0) {
                    if(errno == ENOTDIR) {
                        sleep(1);
                        continue;
                    }
                    fprintf(stderr, "%s: ", prog);
                    perror("stat");
                    return(1);
                }
                if(!S_ISDIR(stbuf.st_mode)) {
                    fprintf(stderr,
                            "%s: %s must be a directory\n", prog, sgidir);
                    return(1);
                }
                printf("%s\n%s\n", sgidir, sgidev);
                return(0);
        }
        fprintf(stderr, "%s: Could not find mounted cdrom\n", prog);
        return(1);
#endif
    }
    else if(volmgt_devcount != 0) {
        fprintf(stderr,
                "%s: You must specify place holders for both the device\n",
                prog);
        fprintf(stderr,
                "and directory name when volume management is being used\n");
        fprintf(stderr, "%s mount ", prog);
        for(i=0;i!= argc;i++)
            fprintf(stderr, "%s ", argv[i]);
        fprintf(stderr, "\n");
        return(1);
    }
/*
 * Either volume management is not running, not managing the devices, or we
 * are not running Solaris 2.x.  We start by checking to see if either the
 * mount point or device are already in use.
 */

    if(argc < 2) {
        fprintf(stderr, "Usage: %s mount device dir\n", prog);
        return(1);
    }

/*
 * Get the absolute path to the mount point, make sure it exists and that
 * the user owns it.  This prevents the user from mounting over / or
 * /usr or other directories.
 */
    if(*(argv[1]) != '/') {
        char junk[MAXPATHLEN+1];
#ifdef __STDC__
        sp = getcwd(junk, MAXPATHLEN);
#else
        sp = getcwd(junk);
#endif
        if(!sp)
            return(1);
/*      tdir = alloca(strlen(sp) + strlen(argv[1]) + 2);*/
        tdir = malloc(strlen(sp) + strlen(argv[1]) + 2);
        sprintf(tdir, "%s/%s", sp, argv[1]);
    }
    else
        tdir = argv[1];

    if(!realpath(tdir, dir)) {
        fprintf(stderr, "%s: realpath ", prog);
        perror(argv[1]);
	free(tdir); /* Added for change of alloca to malloc */
        return(1);
    }

    if(stat(dir, &stbuf) != 0) {
        if(errno != ENOENT) {
            fprintf(stderr, "%s: stat ", prog);
            perror(dir);
	    free(tdir); /* Added for change of alloca to malloc */
            return(1);
        }
        else {
            if(mkdir(dir, 0700) != 0) {
                fprintf(stderr, "%s: mkdir ", prog);
                perror(dir);
		free(tdir); /* Added for change of alloca to malloc */
                return(1);
            }
        }

    }
    else {
        if(!S_ISDIR(stbuf.st_mode)) {
            fprintf(stderr, "%s: %s is not a directory\n", prog, dir);
	    free(tdir); /* Added for change of alloca to malloc */
            return(1);
        }
        if(stbuf.st_uid != olduid && olduid != 0) {
            fprintf(stderr, "%s: You don't own the directory %s\n", prog, dir);
	    free(tdir); /* Added for change of alloca to malloc */
            return(1);
        }
    }

/*
 * Check to see if access is allowed to this device.
 */

    if(!check_device(argv[0])){
	free(tdir); /* Added for change of alloca to malloc */
        return(1);}

    mount_prog = NULL;
/*
 * OSF uses the SVR3 style mount, and calls the filesystem cdfs
 */
#ifdef __osf__
        mount_prog = "/usr/sbin/mount";
        flag = "-t";
        fstype = "cdfs";
#endif
/*
 * This is supposedly the SVR4 style although I haven't checked other
 * SVR4 boxes except Solaris 2.x.
 */
#ifdef __SVR4
        mount_prog = "/usr/sbin/mount";
        flag = "-F";
        fstype = "hsfs";
#else
/*
 * This is the Sun style.  New and old compilers
 */
#if defined(__sun) || defined(sun)
        mount_prog = "/usr/etc/mount";
        flag = "-t";
        fstype = "hsfs";
#endif
#endif

/*
 * IRIX 4.x and 5.x.  IRIX 5.x seems to be like an SVR4 kernel and some apss
 * and the rest is SVR3 plus some BSD.  Just yet another unix flavor.
 */
#ifdef __sgi
        mount_prog = "/usr/sbin/mount";
        flag = "-t";
        fstype = "iso9660";
#endif

if (!mount_prog) {
        fprintf(stderr, "%s: Could not find mount command\n", prog);
	free(tdir); /* Added for change of alloca to malloc */
        return(1);
}

    nargv[0] = "mount";
    nargv[1] = "-r";
    nargv[2] = flag;
    nargv[3] = fstype;
    nargv[4] = argv[0];
    nargv[5] = dir;
    nargv[6] = NULL;

    pid = fork();

    if(pid < 0) {
        fprintf(stderr, "%s: ", prog);
        perror("fork");
	free(tdir); /* Added for change of alloca to malloc */
        return(1);
    }
/*
 * Make ourselves root and mount the device
 */
    if(pid == 0) {
/*
 * Solaris 1.x bug...  /usr/etc must be in your path to run mount and be
 * able to mount file systems.  This is also a security hole since it
 * implies that mount that mount uses exec*p to find mount_hsfs and looks
 * at the user's path to find it.  We must attempt to insure it gets the
 * right mount_hsfs by adding it to the front of the user's path.
 *
 * Also the Sun mount command sends error messages to stdout instead
 * of stderr, so they go back across the pipe when we don't want them
 * too.
 */
#if defined(__sun) || defined(sun)
        sp = getenv("PATH");
        if(!sp){
	   free(tdir); /* Added for change of alloca to malloc */
            return(1);
     }
        nsp = malloc(strlen(sp) + 15);
        strcpy(nsp, "PATH=/usr/etc:");
        strcat(nsp, sp);
        putenv(nsp);

        dup2(2,1);
#endif
        seteuid(0);
        execv(mount_prog, nargv);
        fprintf(stderr, "%s: ", prog);
        perror("execv");
    }
    else {
        wait(&status);
        exit_code = WEXITSTATUS(status);
        if(exit_code != 0){
	   free(tdir); /* Added for change of alloca to malloc */
            return(1);
	 }
    }

    if(verbose)
        printf("%s\n%s\n", argv[1], argv[0]);

    free(tdir); /* Added for change of alloca to malloc */
    return(0);
}

/*
 * In the Solaris 2.x with volume manager we just use eject.  In the
 * no volume management case and under Solaris 1.x, we use umount and
 * then eject.
 */

static int do_unmount(
#ifdef __STDC__
char *prog, char **argv, int argc)
#else
prog, argv, argc)
char *prog, **argv;
int argc;
#endif
{
    char *mount_prog;
    char *nargv[7];
    int exit_code, status;
    pid_t pid;
    struct stat stbuf;

    if(volmgt_devcount != 0)
        goto eject;

/*
 * Check to see if we can use this device
 */

    if(!check_device(argv[0]))
        return(1);

/*
 * Find the unmount command.
 */

    if(stat("/usr/etc/umount", &stbuf) == 0)
        mount_prog = "/usr/etc/umount";
    else if(stat("/usr/sbin/umount", &stbuf) == 0)
        mount_prog = "/usr/sbin/umount";
    else if(stat("/sbin/umount", &stbuf) == 0)
        mount_prog = "/sbin/umount";
    else {
        fprintf(stderr, "%s: Could not find umount command\n", prog);
        return(1);
    }

    nargv[0] = "umount";
    nargv[1] = argv[0];
    nargv[2] = NULL;

    pid = fork();

    if(pid < 0) {
        fprintf(stderr, "%s: ", prog);
        perror("fork");
        return(1);
    }
/*
 * Make ourselves root and unmount the device.
 */
    if(pid == 0) {
        seteuid(0);
        exit_code = execv(mount_prog, nargv);
        fprintf(stderr, "%s: ", prog);
        perror("execv");
        return(exit_code);
    }
    else {
        wait(&status);
        exit_code = WEXITSTATUS(status);
        if(exit_code != 0)
            return(exit_code);
    }

eject:

    if(olduid != -1)
        setuid(olduid);

    if(stat("/usr/bin/eject", &stbuf) == 0)
        mount_prog = "/usr/bin/eject";
    else if(stat("/usr/sbin/eject", &stbuf) == 0)
        mount_prog = "/usr/sbin/eject";
    else {
        fprintf(stderr, "An eject program could not be found\n");
        fprintf(stderr, "Please manually eject the disk\n");
        exit(1);
    }

    execl(mount_prog, "eject", argv[0], NULL);
}

/*
 * OSF does not have a realpath function.  For now we don't actually resolve
 * it, just to get things working.
 */
#if defined(__osf__) || defined(__sgi)
char *realpath(const char *file_name, char *resolved_name)
{
/*    strcpy((char *)file_name, resolved_name); */
    strcpy(resolved_name, (char *)file_name);
    return(resolved_name);
}
#endif
