/******************************************************************************
* C Source: Write Image                                                       
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*    This file contains functions for writing images to disk.                 
*                                                                             
* Module List:                                                                
*    Write_FITS_Extracted_Image                                               
*       Write an image to a FITS file.                                        
*    Write_HHH_Extracted_Image                                                
*       Write an image in GEIS (hhh/hhd) format.                              
*    Write_HHH_Header                                                         
*       Write only the header for an image in GEIS (hhh/hhd) format.          
*    writefits:                                                               
*       Create FITS image file.                                               
*    writehhh:                                                                
*       Create set of GEIS image files.                                       
*                                                                             
* History:                                                                    
*    Created    : 14-DEC-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 29-SEP-95 by J. Doggett,                                    
*       Updated Write_FITS_Extract_Image, Write_HHH_Extracted_Image and       
*       Write_HHH_Header functions.                                           
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
*                                                                             
*    Updated    : 06-NOV-96 by J. Doggett,                                    
*       Added fitsio.h include to support using Pence's CFITSIO package.      
*    Updated    : 22-NOV-96 by J. Doggett,                                    
*       Redelivered Write_FITS_Extracted_Image function.
*    Updated    : 25-NOV-96 by J. Doggett,                                    
*       Redelivered writefits function.
*    Redelivered: xx-xxx-xx by J. Doggett,                                    
*       GetImage Version 2.0x.                                     
*    Removed writing EXTEND keyword (included in new headers) 06-Jul-2005 BJM  
*    Updated to write Mask extension 01-Aug-2005 BJM
*	   GetImage v3.x
*                                                                             
* Copyright (c) 1993-1996, Association of Universities for Research in        
* Astronomy, Inc.  All Rights Reserved.                                       
* Produced under National Aeronautics and Space Administration grant          
* NAG W-2166.                                                                 
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* access function not prototyped in standard place */
#if defined __VMS
#  include <unixio.h>
#elif defined __UNIX
#  include <unixstd.h>
#else
#endif

#include "writeimg.h"

#include "config.h"

/* *** Write Image Introduction ***                                        2 */


#include "extrctn.h"
#include "fitsio.h"
#include "generic.h"
#include "header.h"
#include "image.h"
#include "qfile.h"
#include "writeimg.h"

extern void qclose     ();

/******************************************************************************
* Module:                                                                     
*    Write_FITS_Extracted_Image                                               
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*    Write an extracted image to files in FITS. The file name will be the     
*    root passed in by the user with extension .fits.                         
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Image                                                                 
*          IMAGE_DATA, Reference                                              
*             The image data to write.                                        
*       Header                                                                
*          HEADER                                                             
*             Image header to write with the data.                            
*       imageFileRoot                                                         
*          STRING, Reference                                                  
*             Root for FITS filename                                          
*    Input/Output Arguments:                                                  
*          none                                                               
*    Output Arguments:                                                        
*          none                                                               
*                                                                             
* History:                                                                    
*    Created    : 10-SEP-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 29-SEP-95 by J. Doggett,                                    
*       Removed hard code of file name length.                                
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
*                                                                             
*    Updated    : 22-NOV-96 by J. Doggett,                                    
*        Added calibration file argument which is passed onto writefits.
*    Redelivered: 22-NOV-96 by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
******************************************************************************/
/******************************************
* Write_FITS_Extracted_Image declarations *
********************************************/
void       Write_FITS_Extracted_Image
                 ( Image, Header, imageFileRoot, maskFile, calibrationFile )
IMAGE_DATA Image          ;
HEADER     Header         ;
STRING     imageFileRoot  ;
STRING		maskFile;
STRING     calibrationFile;
{
   char * File_Name               ;
   int    Write_Status            ;
   int    File_Name_Length        ;
   char * File_Extension = ".fits";

/* *** Write_FITS_Extracted_Image ***                                      2 */


/***************************************************
* Write_FITS_Extracted_Image begin executable code *
***************************************************/
   File_Name_Length = strlen(imageFileRoot) + strlen(File_Extension);
   File_Name        = (char *)malloc(File_Name_Length + 1);

   strcpy( File_Name, imageFileRoot );
   strcat( File_Name, File_Extension);

   Write_Status = writefits( Header, Image, File_Name, maskFile, calibrationFile );

   if ( Write_Status != 0 )
      fprintf( stderr, "\nError creating FITS file for %s.\n", imageFileRoot);

   else
      printf( "%s created.\n", File_Name );

   free(File_Name);

   return;
}
/*************************************************
* Write_FITS_Extracted_Image end executable code *
*************************************************/

/******************************************************************************
* Module:                                                                     
*    Write_HHH_Extracted_Image                                                
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*    Write an extracted image to files in hhh/hhd format. The header and data 
*    file names will the root passed in by the user with extensions           
*    .hhh/.hhd.                                                               
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Image                                                                 
*          IMAGE_DATA, Reference                                              
*             The image data to write.                                        
*       Header                                                                
*          HEADER                                                             
*             Image header to write with the data.                            
*       imageFileRoot                                                         
*          STRING, Reference                                                  
*             Root for FITS filename                                          
*    Input/Output Arguments:                                                  
*          none                                                               
*    Output Arguments:                                                        
*          none                                                               
*                                                                             
* History:                                                                    
*    Created    : 10-SEP-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 29-SEP-95 by J. Doggett,                                    
*       Removed hard code of file name length.                                
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
******************************************************************************/
/*****************************************
* Write_HHH_Extracted_Image declarations *
*****************************************/
void       Write_HHH_Extracted_Image( Image, Header, imageFileRoot )
IMAGE_DATA Image        ;
HEADER     Header       ;
STRING     imageFileRoot;
{
   char * File_Name              ;
   int    Write_Status           ;
   int    File_Name_Length       ;
   char * File_Extension = ".hhh";

/**************************************************
* Write_HHH_Extracted_Image begin executable code *
**************************************************/
   File_Name_Length = strlen(imageFileRoot) + strlen(File_Extension);
   File_Name = (char *)malloc(File_Name_Length + 1);

/* *** Write_HHH_Extracted_Image ***                                       2 */


   strcpy( File_Name, imageFileRoot );
   strcat( File_Name, File_Extension);

   Write_Status = writehhh( Header, Image, File_Name );

   if ( Write_Status != 0 )
      fprintf( stderr, "\nError creating GEIS file for %s.\n", imageFileRoot);
   else
      printf( "%s/hhd created.\n", File_Name );

   free(File_Name);

   return;
}
/************************************************
* Write_HHH_Extracted_Image end executable code *
************************************************/

/******************************************************************************
* Module:                                                                     
*    Write_HHH_Header                                                         
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*    Write an hhh header file for an image.  The file name will be the root   
*    passed in by the user with extension .hhh.                               
*                                                                             
* Return Value:                                                               
*    none                                                                     
*                                                                             
* Arguments;                                                                  
*                                                                             
*    Input Arguments:                                                         
*       Header                                                                
*          HEADER                                                             
*             Image header to write with the data.                            
*       imageFileRoot                                                         
*          STRING, Reference                                                  
*             Root for FITS filename                                          
*    Input/Output Arguments:                                                  
*          none                                                               
*    Output Arguments:                                                        
*          none                                                               
*                                                                             
* History:                                                                    
*    Created    : 07-DEC-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 29-SEP-95 by J. Doggett,                                    
*       Removed hard code of file name length.                                
*    Redelivered: 01-AUG-96 by J. Doggett,                                    
*       GetImage Version 1.3.                                                 
******************************************************************************/
/********************************
* Write_HHH_Header declarations *
********************************/
void       Write_HHH_Header(Header, imageFileRoot)
HEADER     Header       ;
STRING     imageFileRoot;
{
   char  * File_Name              ;
   QFILE * ihead                  ;
   int     File_Name_Length       ;
   char  * File_Extension = ".hhh";

   int i;

/*****************************************
* Write_HHH_Header begin executable code *
*****************************************/
   File_Name_Length = strlen(imageFileRoot) + strlen(File_Extension);
   File_Name = (char *)malloc(File_Name_Length + 1);

   strcpy( File_Name, imageFileRoot );
   strcat( File_Name, File_Extension);

/* *** Write_HHH_Header ***                                                2 */


    /*
     * write header file: 80 byte records, CR carriage control
     */
    ihead = qcreat(File_Name,80,1);
    for (i=0; i < Header->nhlines; i++) qwrite(ihead,Header->hlines[i],80);

    qclose(ihead);
    printf( "%s created.\n", File_Name );

    free(File_Name);

   return;
}
/***************************************
* Write_HHH_Header end executable code *
***************************************/

/******************************************************************************
* Module:                                                                     
*    writefits                                                                
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*       Write FITS format data file                                           
*                                                                             
* Return Value:                                                               
*    int: 0 if successful, -1 if fails.                                       
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input Arguments:                                                         
*       h                                                                     
*          HEADER, Reference                                                  
*             Header structure, contains proper image size updated with       
*             fillhdr.                                                        
*       image                                                                 
*          short int, Reference                                               
*             The image data to write.                                        
*       headfile                                                              
*          char, Reference                                                    
*             Name of the hhh header file. Must end in .hhh. The final h is   
*             changed to a d for the data file.                               
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 31-JUL-91 by R. White,                                      
*    Delivered  : 16-JUN-92 by R. White,                                      
*        Modified to write OpenVMS fixed-length record.                       
*                                                                             
*    Updated    : 10-SEP-93 by J. Doggett,                                    
*        Adopted into software used to facilitate the accessing of data from  
*            the CD-ROM set of the Digitized Sky Survey.                      
*    Updated    : 06-JAN-94 by J. Doggett,                                    
*       Changed #if defs to check for UNIX rather not OpenVMS when checking   
*          if we might be writing over a file.                                
*    Updated    : 10-JAN-94 by J. Doggett,                                    
*       Modified #if def to abort if we're not on a known operating system.   
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 06-NOV-96 by J. Doggett,                                    
*       Re-wrote to use Pence's CFITSIO.  This required no longer supporting
*       making new versions under VMS because CFITSIO does not support it.
*    Updated    : 22-NOV-96 by J. Doggett,                                    
*       Modified to accept calibration file argument and to copy calibration
*       FITS extension from calibration file into output image FITS file.
*    Updated    : 25-NOV-96 by J. Doggett,                                    
*       Modified to include EXTEND keyword.
*    Updated    : 12-DEC-96 by J. Doggett,                                    
*       Modified to create image FITS file even if the calibration FITS file
*       is not accessible.
*    Redelivered: xx-xxx-xx by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
******************************************************************************/
extern int
writefits(h,image,filename,maskFileName, calibrationFileName)

/* *** writefits ***                                                       2 */


HEADER h;
short image[];
char *filename;
char *maskFileName;
char *calibrationFileName;
{
fitsfile * fits_file       ;
fitsfile * mask_file;
fitsfile * calibration_file;
int        status          ;
int        i               ;
int        datatype        ;
long       firstelem       ;
long       nelements       ;
int        hdutype         ;
char       NAXISstring[81] ;
char       NAXIScomment[81];
int        scanResult      ;
int        NAXISvalue      ;
int        nthAxis         ;
char       NAXISnKeyword[9];
int        NAXIS           ;
int	extCount;

    /*
     * Check to see whether file already exists
     * (Don't bother under OpenVMS, just make a new version)
     */
   if (access(filename,0) == 0) 
	{
        fprintf(stderr, "Error: file %s already exists\n",filename);
        return(-1);
    }

   status = 0;
   fits_create_file(&fits_file, filename, &status);
   if (status != 0)
	{
         fprintf(stderr, "Error opening FITS file %s\n", filename);
         return(-1);
	}


/* Check SIMPLE */
   if (strncmp(h->hlines[0], "SIMPLE  ", 8) != 0)
	{
		fprintf(stderr, "SIMPLE keyword misplaced.\n");
		status = 0;
		fits_delete_file(fits_file, &status);
		if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		return -1;
	}
   status = 0;
   fits_write_record(fits_file, h->hlines[0], &status);
   if (status != 0)
   {
		fprintf(stderr, "Error writing SIMPLE in FITS header.\n");
		status = 0;
		fits_delete_file(fits_file, &status);
		if (status != 0) 
	    fprintf(stderr, "Error deleting FITS file %s\n", filename);
		return(-1);
	}

/* Check BITPIX */
   if (strncmp(h->hlines[1], "BITPIX  ", 8) != 0)
   {
	   fprintf(stderr, "BITPIX keyword misplaced.\n");
	   status = 0;
	   fits_delete_file(fits_file, &status);
	   if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
	   return -1;
	}
   status = 0;
   fits_write_record(fits_file, h->hlines[1], &status);
   if (status != 0)
   {
	   fprintf(stderr, "Error writing BITPIX in FITS header.\n");
	   status = 0;
	   fits_delete_file(fits_file, &status);
	   if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
	   return(-1);
	}

/* Check NAXIS */
   if (strncmp(h->hlines[2], "NAXIS   ", 8) != 0)
	{
         fprintf(stderr, "NAXIS keyword misplaced.\n");
	   status = 0;
	   fits_delete_file(fits_file, &status);
	   if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
	   return -1;
	}
   status = 0;
   fits_write_record(fits_file, h->hlines[2], &status);
	if (status != 0)
	{
		fprintf(stderr, "Error writing NAXIS in FITS header.\n");
		status = 0;
		fits_delete_file(fits_file, &status);
		if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		return(-1);
	}

/* Check NAXISn */
   status = 0;
   fits_parse_value(h->hlines[2], NAXISstring, NAXIScomment, &status);
	if (status != 0)
	{
		fprintf(stderr, "Error parsing NAXIS.\n");
		status = 0;
		fits_delete_file(fits_file, &status);
		if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		return(-1);
	}
    scanResult = sscanf(NAXISstring, "%d", &NAXISvalue);
	if (scanResult != 1)
	{
		fprintf(stderr, "Error parsing NAXIS.\n");
		status = 0;
		fits_delete_file(fits_file, &status);
	    if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		return(-1);
	}

   for (nthAxis = 1; nthAxis <= NAXISvalue; nthAxis++)
   {
      status = 0;
      fits_make_keyn("NAXIS", nthAxis, NAXISnKeyword, &status);
	  if (status != 0)
	  {
		  fprintf(stderr, "Error parsing NAXIS.\n");
		  status = 0;
		  fits_delete_file(fits_file, &status);
		  if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		  return(-1);
	  }
	  if (strncmp(h->hlines[2 + nthAxis], NAXISnKeyword, strlen(NAXISnKeyword)) != 0)
	  {
		  fprintf(stderr, "NAXISn keyword misplaced.\n");
		  status = 0;
		  fits_delete_file(fits_file, &status);
		  if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		  return -1;
	  }
   
      status = 0;
      fits_write_record(fits_file, h->hlines[2 + nthAxis], &status);
	  if (status != 0)
	  {
		  fprintf(stderr, "Error writing NAXISn in FITS header.\n");
		  status = 0;
		  fits_delete_file(fits_file, &status);
		  if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		  return(-1);
	  }
   }

/* Write EXTEND
   status = 0;
   fits_write_key_log(fits_file, "EXTEND  ", (int)1, " ", &status);
      if (status != 0)
      {
         fprintf(stderr, "Error writing EXTEND in FITS file.\n");
	 status = 0;
	 fits_delete_file(fits_file, &status);
	    if (status != 0) 
	       fprintf(stderr, "Error deleting FITS file %s\n", filename);

	 return(-1);
      } */

/* Write rest of header */
   for (i = 3 + NAXISvalue; i < h->nhlines - 1; i++) /* nhlines-1 because don't write END */
   {
	   fits_write_record(fits_file, h->hlines[i], &status);
	   if (status != 0)
	   {
		   fprintf(stderr, "Error writing FITS file header %s\n", filename);
		   status = 0;
		   fits_delete_file(fits_file, &status);
		   if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);
		   return(-1);
	   }
   }

/* Write data */
   datatype  = TSHORT;
   firstelem = 1     ;
   nelements = h->section_x_length * h->section_y_length;
   fits_write_img(fits_file, datatype, firstelem, nelements, image, &status);
   if (status != 0)
   {
	   fprintf(stderr, "Error writing FITS file data %s\n", filename);
	   status = 0;
	   fits_delete_file(fits_file, &status);
	   if (status != 0) fprintf(stderr, "Error deleting FITS file %s\n", filename);		
	   return(-1);
	}

	status = 0;
	fits_close_file(fits_file, &status);
	if (status != 0) 
	{
		fprintf(stderr, "#1 Error closing FITS file %s\n", filename);
		return(-1);
	}

	/* Write table extensions*/
	/* Open FITS file for writing */
	status = 0;
	fits_open_file(&fits_file, filename, READWRITE, &status);
	if (status != 0){
		fits_clear_errmsg();
		printf("Error opening FITS file %s;\n",maskFileName);
		return (-1);
	}
	else {
		/* Get mask data */
		extCount=0;
		if (maskFileName !=(char *)NULL){
			fits_open_file(&mask_file, maskFileName, READONLY, &status);
			if (status != 0){
				fits_clear_errmsg();
				printf("Error opening mask FITS file %s;\n",maskFileName);
				printf("\tMask data not written to FITS file.\n");
			} 
			else {
				fits_movrel_hdu(mask_file, 1, &hdutype, &status);
				fits_create_hdu(fits_file, &status);
				fits_copy_hdu(mask_file, fits_file, 0, &status);
				fits_movrel_hdu(fits_file,1,&hdutype, &status);
				fits_close_file(mask_file, &status);
			}
		}
	
		/* Get Calibration Data */			
		if (calibrationFileName !=(char *)NULL){
			status=0;
			fits_open_file(&calibration_file, calibrationFileName, READONLY, &status);
			if (status != 0){
				fits_clear_errmsg();
				printf("Error opening mask clibration file %s;\n",calibrationFileName);
				printf("\tCalibration data not written to FITS file.\n");
			}
			else {
				fits_movrel_hdu(calibration_file, 1, &hdutype, &status);
				fits_create_hdu(fits_file, &status);
				fits_copy_hdu(calibration_file, fits_file, 0, &status);
				fits_movrel_hdu(fits_file,1,&hdutype, &status);
				fits_close_file(calibration_file, &status);
			}
		} 
	}
	fits_close_file(fits_file, &status);
	return(0);
}
/******************************************************************************
* Module:                                                                     
*    writehhh                                                                 
*                                                                             
* File:                                                                       
*    writeimg.c                                                               
*                                                                             
* Description:                                                                
*       Write hhh/hhd format data files.                                      
*                                                                             
* Return Value:                                                               
*    int: 0 if successful, -1 if fails.                                       
*                                                                             
* Arguments:                                                                  
*                                                                             
*    Input Arguments:                                                         
*       h                                                                     
*          HEADER, Reference                                                  
*             Header structure, contains proper image size updated with       
*             fillhdr.                                                        
*       image                                                                 
*          short int, Reference                                               
*             The image data to write.                                        
*       headfile                                                              
*          char, Reference                                                    
*             Name of the hhh header file. Must end in .hhh. The final h is   
*             changed to a d for the data file.                               
*    Input/Output Arguments:                                                  
*       none                                                                  
*    Output Arguments:                                                        
*       none                                                                  
*                                                                             
* History:                                                                    
*    Created    : 31-JUL-91 by R. White,                                      
*    Delivered  : 16-JUN-92 by R. White,                                      
*                  Modified to write OpenVMS fixed-length record.             
*                                                                             
*    Updated    : 10-SEP-93 by J. Doggett,                                    
*        Adopted into software used to facilitate the accessing of data from  
*            the CD-ROM set of the Digitized Sky Survey.                      
*    Updated    : 06-JAN-94 by J. Doggett,                                    
*       Changed #if defs to check for UNIX rather not OpenVMS when checking   
*          if we might be writing over a file.                                
*    Updated    : 10-JAN-94 by J. Doggett,                                    
*       Modified #if def to abort if we're not on a known operating system.   
*    Redelivered: 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
******************************************************************************/
extern int
writehhh(h,image,headfile)
/*Header *h;*/
HEADER h;
short image[];
char *headfile;
{
char *datafile, *p;
int i, npix;
QFILE *ihead, *idata;

    /*
     * construct filename for data (hhd) file
     */
    datafile = malloc(strlen(headfile)+1);
    strcpy(datafile, headfile);

/* *** writehhh ***                                                        2 */


    p = strrchr(datafile, 'h');
    *p++ = 'd';

    /*
     * the 'h' must either be the last character or be followed by
     * a semicolon (to be compatible with OpenVMS)
     */
    if (*p != '\0' && *p != ';') {
        fprintf(stderr, "Error: header filename must end with 'h'\n");
        free(datafile);
        return(-1);
    }
    /*
     * check to see whether header or data files already exist
     * (Don't bother under OpenVMS, just make a new version)
     */

    if (access(headfile,0) == 0) {
#if defined __VMS
/* We'll just make a new version */
#elif defined _UNIX
         fprintf(stderr, "Error: header file %s already exists\n",headfile);
         free(datafile);
         return(-1);
#else
         fprintf(stderr,"Unknown file system error in writehhh");
         exit(EXIT_FAILURE);
#endif
    }

    if (access(datafile,0) == 0) {
#if defined __VMS
/* We'll just make a new version */
#elif defined _UNIX
        fprintf(stderr, "Error: data file %s already exists\n",datafile);
        free(datafile);
        return(-1);
#else
        fprintf(stderr,"Unknown file system error in writehhh");
        exit(EXIT_FAILURE);
#endif
    }

    /*
     * write header file: 80 byte records, CR carriage control
     */
    ihead = qcreat(headfile,80,1);
    for (i=0; i < h->nhlines; i++) qwrite(ihead,h->hlines[i],80);
    qclose(ihead);
    /*
     * write data file: 512 byte records, no carriage control
     */
    idata = qcreat(datafile,512,0);
    npix = h->section_x_length * h->section_y_length;
    qwrite(idata, image, sizeof(short)*npix);
    qclose(idata);
    free(datafile);
    return(0);

} /* writehhh */
